"""Resource action method decorators specific for the 'byoml' service."""

from functools import wraps
import logging
from tenacity import (
    Retrying,
    RetryCallState,
    stop_after_attempt,
    stop_after_delay,
    wait_exponential,
    retry_if_exception
)

from simple_rest_client.exceptions import (
    ErrorWithResponse, ClientConnectionError
)

from waylay.exceptions import RestConnectionError
from waylay.service._base import WaylayAction
from ._exceptions import ByomlActionError, ByomlActionParseError, ModelNotReadyError

DEFAULT_BYOML_MODEL_TIMEOUT = 60
TEMPORARY_FAILURES = [
    409,
    429,
    500,
    502,
    503,
    504,
    508
]
TEMPORARY_FAILURE_EXCEPTIONS = [
    ModelNotReadyError
]
WAIT_EXPONENTIAL = {
    "multiplier": 1,
    "min": 4,
    "max": 32
}
DEFAULT_RETRY_ATTEMPTS = 20
DEFAULT_RETRY_MAX_DELAY = 180


def before_sleep_logger(action: WaylayAction, logger=None, level=logging.WARNING):
    """Create a before-sleep log function for tenacity.Retrying."""
    logger = logger or logging.getLogger(action.fqn)

    def _log_before_sleep(retry_state: RetryCallState) -> None:
        next_action = retry_state.next_action
        outcome = retry_state.outcome
        assert next_action and outcome
        logger.log(
            level,
            "Retrying %s in %s seconds as it raised %s.",
            action.fqn,
            next_action.sleep,
            outcome.exception()
        )
    return _log_before_sleep


def byoml_exception_decorator(action_method):
    """Create a decorator that parses json error responses."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except ErrorWithResponse as exc:
            raise ByomlActionError.from_cause(exc) from exc
        except ClientConnectionError as exc:
            raise RestConnectionError.from_cause(exc) from exc
    return wrapped


def byoml_raise_not_ready_get(action_method):
    """Create a decorator that retries after certain exceptions."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        retry_until_ready = kwargs.pop('retry_until_ready', False)
        response = action_method(*args, **kwargs)
        if retry_until_ready:
            is_ready = response.body.get('ready', None)
            if is_ready is None:
                raise ByomlActionParseError('Failed to extract `ready` attribute', response)
            if not is_ready:
                raise ModelNotReadyError('Model is not ready yet.')
        return response
    retry_until_ready_arg = f'retry_until_ready=False'
    wrapped.__doc__ = ((wrapped.__doc__ or '') + f"""
The response contains a 'ready' attribute that indicates whether the model is ready to serve.
    {retry_until_ready_arg:30}: If true, raise an error handled by the retry mechanism when not ready.
""")
    return wrapped


def byoml_retry_upload_after_deletion_decorator(action_method):
    """
    Create a decorator that retries uploading a model.

    Occurs after a model with the same name has been deleted.
    """

    @wraps(action_method)
    def wrapped(*args, **kwargs):
        before_sleep = before_sleep_logger(action_method.action)
        retry_max_delay = kwargs.pop('retry_max_delay', DEFAULT_RETRY_MAX_DELAY)
        # if timeout is not explicitely given, align with retry_max_delay
        timeout = kwargs.pop('timeout', retry_max_delay)
        for attempt in Retrying(
            stop=(stop_after_delay(retry_max_delay)),
            retry=retry_if_exception(is_409_error),
            before_sleep=before_sleep,
            wait=wait_exponential(**WAIT_EXPONENTIAL),
            reraise=True
        ):
            with attempt:
                # reset input stream when retrying upload
                body = kwargs.get('body', None)
                if body and hasattr(body, 'seek'):
                    body.seek(0)
                return action_method(*args, timeout=timeout, **kwargs)

    retry_max_delay_arg = f'retry_max_delay={DEFAULT_RETRY_MAX_DELAY}'
    wrapped.__doc__ = ((wrapped.__doc__ or '') + f"""
Retries on upload after a deletion of a model with the same name has occurred (HTTP status code: 409).
Arguments that configure retry:
    {retry_max_delay_arg:30}: Maximal delay in seconds.
""")
    return wrapped


def byoml_retry_decorator(action_method):
    """Create a decorator that retries after certain exceptions."""

    @wraps(action_method)
    def wrapped(*args, **kwargs):
        before_sleep = before_sleep_logger(action_method.action)
        retry_attempts = kwargs.pop('retry_attempts', DEFAULT_RETRY_ATTEMPTS)
        retry_max_delay = kwargs.pop('retry_max_delay', DEFAULT_RETRY_MAX_DELAY)
        # if timeout is not explicitely given, align with retry_max_delay
        timeout = kwargs.pop('timeout', retry_max_delay)
        for attempt in Retrying(
            stop=(stop_after_attempt(retry_attempts) | stop_after_delay(retry_max_delay)),
            retry=retry_if_exception(is_retry_failure),
            before_sleep=before_sleep,
            wait=wait_exponential(**WAIT_EXPONENTIAL),
            reraise=True
        ):
            with attempt:
                return action_method(*args, timeout=timeout, **kwargs)

    retry_attempts_arg = f'retry_attempts={DEFAULT_RETRY_ATTEMPTS}'
    retry_max_delay_arg = f'retry_max_delay={DEFAULT_RETRY_MAX_DELAY}'
    wrapped.__doc__ = ((wrapped.__doc__ or '') + f"""
Retries on temporary failures (HTTP status codes {', '.join(str(c) for c in TEMPORARY_FAILURES)}).
Arguments that configure retry:
    {retry_attempts_arg:30}: Maximal number of retries.
    {retry_max_delay_arg:30}: Maximal delay in seconds.
""")
    return wrapped


def is_retry_failure(exc):
    """Check if exception is a custom error or temporary error."""
    return is_temporary_failure_exception(exc) or is_temporary_failure(exc)


def is_temporary_failure_exception(exc):
    """Check if a given exception is a custom temporary failure exception."""
    return type(exc) in TEMPORARY_FAILURE_EXCEPTIONS


def is_temporary_failure(exc):
    """Check if given exception is temporary."""
    return hasattr(exc, "response") and exc.response.status_code in TEMPORARY_FAILURES


def is_409_error(exc):
    """Check if given exception is a 409 error."""
    return hasattr(exc, 'response') and exc.response.status_code == 409



"""REST definitions for the 'runtime' entity of the 'byoml' service."""
from typing import (
    Optional, Dict
)
from .._base import WaylayResource
from .._decorators import (
    return_body_decorator,
    return_path_decorator,
)
from ._decorators import (
    byoml_exception_decorator,
    byoml_retry_decorator,
)

FRAMEWORK_ARG = {
    'name': 'framework',
    'type': 'str',
    'description': 'Byoml framework name for a supported runtime.',
    'examples': ['sklearn', 'custom']
}
FRAMEWORK_VERSION_ARG = {
    'name': 'framework_version',
    'type': 'str',
    'description': (
        'Framework version specifier. '
        'Supports (python version specifiers)[https://packaging.pypa.io/en/latest/specifiers.html]'
    ),
    'examples': ['2.1', '>=2.1,<2.4']
}
FRAMEWORK_LIST_RETURN = {
    'name': 'frameworks',
    'type': 'List[Dict]',
    'description': 'Representation of supported frameworks.'
}


class FrameworkResource(WaylayResource):
    """REST Inspect supported runtimes for a given framework."""

    link_roots = {
        'doc': '${doc_url}/api/byoml/?id=',
        'apidoc': '${apidoc_url}/byoml.html'
    }

    actions = {
        'list': {
            'method': 'GET',
            'url': '/frameworks',
            'returns': [
                FRAMEWORK_LIST_RETURN
            ],
            'decorators': [
                byoml_exception_decorator,
                return_path_decorator(['frameworks'])
            ],
            'description': 'Frameworks and supported runtimes.',
            'links': {
                'doc': 'runtimes',
                'apidoc': '',
            },
        },
        'get': {
            'method': 'GET',
            'url': '/frameworks/{}',
            'arguments': [FRAMEWORK_ARG],
            'decorators': [
                byoml_exception_decorator,
                byoml_retry_decorator,
                return_body_decorator,
            ],
            'description': 'Get the default runtime for a framework.',
            'links': {
                'doc': 'runtimes',
                'apidoc': '',
            },
        },
        'list_versions': {
            'method': 'GET',
            'url': '/frameworks/{}/versions',
            'arguments': [FRAMEWORK_ARG],
            'decorators': [
                byoml_exception_decorator,
                byoml_retry_decorator,
                return_body_decorator,
            ],
            'description': 'Get the runtimes for a framework.',
            'links': {
                'doc': 'runtimes',
                'apidoc': '',
            },
        },
        'get_version': {
            'method': 'GET',
            'url': '/frameworks/{}/versions/{}',
            'arguments': [FRAMEWORK_ARG, FRAMEWORK_VERSION_ARG],
            'decorators': [
                byoml_exception_decorator,
                byoml_retry_decorator,
                return_body_decorator,
            ],
            'description': 'Get the runtime for a given framework and framework version.',
            'links': {
                'doc': 'runtimes',
                'apidoc': '',
            },
        }
    }

    def find_runtime(self, framework: str, framework_version: Optional[str]) -> Dict:
        """Get the byoml plug runtime for this framework and version (default version if not specified).

        framework: the name of the framework
        framework_version: (optional) the version for this framework, the default version is defined by
                           the byoml service itself.
        """
        if framework_version:
            return self.get_version(framework, framework_version)  # pylint: disable=no-member
        return self.get(framework)['default_runtime']  # pylint: disable=no-member



"""Exceptions specific to the Byoml Service."""

from ...exceptions import (
    RestResponseError,
    RestResponseParseError,
    RequestError,
    RestError
)


class ByomlActionError(RestResponseError):
    """Error that represents the json messages of a byoml response."""

    @property
    def message(self):
        """Get the main user error returned by a byoml error response."""
        return self._get_from_body('error', super().message)


class ByomlActionParseError(RestResponseParseError, ByomlActionError):
    """Indicates that a byoml response could not be parsed."""


class ByomlValidationError(RequestError):
    """Exception class for BYOML validation errors."""


class ModelNotReadyError(RestError):
    """Indicates that a byoml action should be retried."""



"""REST definitions for the 'series' entity of the 'data' service."""

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]

RESOURCE_ARG = {
    'name': 'resource_id',
    'type': 'str',
    'description': 'The id of a waylay resource.'
}


class EventsResource(WaylayResource):
    """REST Resource for the 'events' ingestion of the 'data' service."""

    link_roots = {
        'doc': '${doc_url}/api/broker/?id=',
        'apidoc': '${apidoc_url}/broker.html'
    }

    actions = {
        'post': {
            'method': 'POST', 'url': '/events/{}',
            'arguments': [RESOURCE_ARG],
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator([]),
            ],
            'description': (
                'Forward a json message to the rule engine, '
                'time series database and/or document store for a given resource.'
            ),
            'links': {
                'doc': 'posting-data-to-the-storage-and-rule-engine',
                'apidoc': '',
            },
        },
        'bulk': {
            'method': 'POST', 'url': '/events',
            'returns': [
            ],
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator([])
            ],
            'description': (
                'Forward an array of json messages to the rule engine, '
                'time series database and/or document store.'
            ),
            'links': {
                'doc': 'posting-array-of-data',
                'apidoc': '',
            },
        },
        'remove': {
            'method': 'DELETE', 'url': '/{}',
            'arguments': [RESOURCE_ARG],
            'decorators': DEFAULT_DECORATORS,
            'description': 'Remove all data for a resource.',
            'links': {
                'doc': 'all-data-for-a-resource',
                'apidoc': '',
            },
        },
    }

    def get_action_full_url(self, action_name, *parts):
        """Override the regular url computation when not using api gateway."""
        via_gateway = '/data/v1' in self.api_root_url
        if not via_gateway and action_name == 'remove':
            # remove not via api:
            return self.api_root_url + f'resources/{parts[0]}'
        else:
            return super().get_action_full_url(action_name, *parts)



"""REST definitions for the 'series' entity of the 'data' service."""
from typing import Any, Dict, Iterator, Tuple, Optional

import urllib.parse

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]
DEFAULT_EXPORT_PAGE_SIZE = 2000


RESOURCE_ARG = {
    'name': 'resource_id',
    'type': 'str',
    'description': 'The id of a waylay resource.'
}
METRIC_ARG = {
    'name': 'metric',
    'type': 'str',
    'description': 'The name of the metric of a resource series.'
}
ADDITIONAL_PARAMS_ARG = {
    'name': 'params',
    'type': 'dict',
    'description': 'Additional parameters, mapped to url query parameters. See API documentation.'
}


class SeriesResource(WaylayResource):
    """REST Resource for the 'series' entity of the 'data' service."""

    link_roots = {
        'doc': '${doc_url}/api/broker/?id=',
        'apidoc': '${apidoc_url}/broker.html'
    }

    actions = {
        'data': {
            'method': 'GET', 'url': '/series/{}/{}',
            'arguments': [
                RESOURCE_ARG, METRIC_ARG, ADDITIONAL_PARAMS_ARG
            ],
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['series']),
            ],
            'description': 'Retrieve the (optionally aggregated) data of a single series.',
            'links': {
                'doc':  'getting-time-series-data'
            },
        },
        'list': {
            'method': 'GET', 'url': '/series/{}',
            'arguments': [RESOURCE_ARG, ADDITIONAL_PARAMS_ARG],
            'decorators': [
                decorators.default_params_decorator({'metadata': 'true'}),
                decorators.exception_decorator,
                decorators.return_path_decorator([]),
            ],
            'description': 'Retrieve a list of series and their latest value for a given resource.',
            'links': {
                'doc': 'metadata',
                'apidoc': '',
            },
        },
        'latest': {
            'method': 'GET', 'url': '/series/{}/{}/latest',
            'arguments': [RESOURCE_ARG, METRIC_ARG],
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator([])
            ],
            'description': 'Fetch the latest value for a series.',
            'links': {
                'doc': 'latest-value-for-a-series',
                'apidoc': '',
            },
        },
        'query': {
            'method': 'POST', 'url': '/series/query',
            'arguments': [{
                'name': 'body',
                'type': 'Dict',
                'description': 'A series query specification (see REST API doc).'
            }],
            'description': 'Execute a broker query document to retrieve aggregated timeseries.',
            'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['series'])
            ],
            'links': {
                'doc': 'post-timeseries-query',
                'apidoc': '',
            },
        },
        'export': {
            'method': 'GET',
            'url': '/series/{}/{}/raw',
            'arguments': [RESOURCE_ARG, METRIC_ARG, ADDITIONAL_PARAMS_ARG],
            'description': 'Export a single series using paging with HAL links.',
            'decorators': [
                decorators.exception_decorator,
                decorators.default_params_decorator(
                    {'from': 0, 'limit': DEFAULT_EXPORT_PAGE_SIZE, 'order': 'ascending'}),
                decorators.default_header_decorator({'Accept': 'application/hal+json'}),
                decorators.return_path_decorator(['series']),
            ],
            'links': {
                'doc': 'getting-time-series-data',
                'apidoc': '',
            },
        }
    }

    def iter_export(
        self, resource_id: str, metric: str,
        page_size: Optional[int] = None,
        descending: bool = False,
        params: Optional[Dict[str, Any]] = None,
        **kwargs: Any
    ) -> Iterator[Tuple[int, Any]]:
        """Fully export a timeseries.

        Uses paging on the '/series/{}/{}/raw' endpoint.

        Parameters
            resource_id: the resource id for the timeseries
            metric: the metric of the timeseries
            page_size: the `limit` size with which data is retrieved using the `export` action
            descending: if true, the series is iterated descending (`order=descending`)
            params: URL params that are passed to the `export` action. These might include:
                params.from: export start timestamp
                params.until: export end timestamp
            kwargs: other params are passed onto the `export` action
        """
        export_params = params or {}
        export_params['limit'] = page_size
        if descending:
            export_params['order'] = 'descending'
        if kwargs.pop('select_path', None):
            raise AttributeError('Cannot specify `select_path` in `iter_export`')
        while True:
            resp = self.export(  # pylint: disable=no-member
                resource_id, metric,
                select_path=None, params=export_params, **kwargs
            )
            if 'series' in resp:
                yield from resp['series']
            next_link = resp.get('_links', {}).get('next', {}).get('href')
            if not next_link:
                return
            next_query = urllib.parse.urlparse(next_link).query
            export_params = urllib.parse.parse_qs(next_query)




"""REST client for the Waylay Queries Service."""

from .._base import WaylayRESTService

from .query import QueryResource
from .about import AboutResource


class QueriesService(WaylayRESTService):
    """REST client for the Waylay Queries Service."""

    service_key = 'queries'
    config_key = 'query'
    gateway_root_path = '/queries/v1'
    default_root_path = '/queries/v1'

    resource_definitions = {
        'query': QueryResource,
        'about': AboutResource
    }
    query: QueryResource
    about: AboutResource

    def __init__(self):
        """Create a QueriesService."""
        super().__init__(json_encode_body=True)



"""Legacy placeholder for the Analytics REST service."""


class AnalyticsServiceLegacy:
    """Legacy placeholder."""

    @property
    def query(self):
        """Raise a unsupported issues."""
        raise NotImplementedError(
            'The `analytics.query` resource is no longer supported. Please use `queries.query`.'
        )

    @property
    def about(self):
        """Raise a unsupported issues."""
        raise NotImplementedError(
            'The `analytics` service is no longer supported. Please use `queries` service.'
        )



"""Command line interface for the Waylay Python SDK."""

import argparse
import logging

from waylay import WaylayClient

from . import seriescli, servicecli, configcli


CLI_COMMANDS = [CMD_SERVICE, CMD_SERIES, CMD_CONFIG] = ['service', 'series', 'config']


def main():
    """Start the waylaycli program."""
    logging.basicConfig()
    parser = argparse.ArgumentParser(
        prog='waylaycli', description='Command line interface to the Waylay Python SDK',
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        '-p', '--profile', dest='profile', nargs='?', help='Configuration profile.', default=None
    )
    parser.add_argument(
        '-l', '--loglevel', dest='log_level', nargs='?',
        default='WARNING', type=lambda _: _.upper(),
        help=f"Log level for waylay packages. One of {','.join(logging._levelToName.values())}.",

    )
    parser.add_argument(
        '--libloglevel', dest='lib_log_level', nargs='?',
        default='WARNING', type=lambda _: _.upper(),
        help=f"Log level for other packages. One of {','.join(logging._levelToName.values())}.",
    )
    cmd_parsers = parser.add_subparsers(dest='cmd')
    servicecli.init_srv_parser(
        cmd_parsers.add_parser(CMD_SERVICE, help='List available services.'))
    seriescli.init_series_parser(
        cmd_parsers.add_parser(CMD_SERIES, help='Interact with waylay timeseries.'))
    configcli.init_parser(
        cmd_parsers.add_parser(CMD_CONFIG, help='Manage Waylay client configuration profiles.'))
    args = parser.parse_args()

    def waylay_client():
        return WaylayClient.from_profile(args.profile)

    waylay_logger = logging.getLogger('waylay')
    try:
        import coloredlogs
        coloredlogs.install(level='DEBUG')
    except ImportError:
        pass
    logging.getLogger().setLevel(args.lib_log_level)
    waylay_logger.setLevel(args.log_level)
    done = False
    if args.cmd == CMD_SERVICE:
        done = servicecli.handle_srv_cmd(waylay_client(), args)
    if args.cmd == CMD_SERIES:
        done = seriescli.handle_series_cmd(waylay_client(), args)
    if args.cmd == CMD_CONFIG:
        done = configcli.handle_cmd(None, args)
    if not done:
        parser.print_help()



"""Python SDK Documentation tooling."""

from typing import Optional, Dict, Union, List, Any, cast

import pandas as pd
import numpy as np

from waylay.service import WaylayResource, WaylayServiceContext, WaylayAction

DESC_LEVELS = [
    DESC_DESC,
    DESC_ARGS,
    DESC_RETURN,
    DESC_REST,
    DESC_WRAPPED,
    DESC_LINK,
    DESC_SDK_ACTION
] = [
    'title',
    'arguments',
    'return',
    'rest',
    'wrapped',
    'link',
    'sdk'
]


def format_links_html(links: Optional[Dict[str, Dict[str, str]]]) -> str:
    """Format link information to an html string."""
    if not links:
        return ''
    link_items = (
        f'<a href="{href}" target="{target}">{text}</a>'
        for target, href, text in (
            (f"_{key}", link.get('href'), link.get('title', key))
            for key, link in links.items()
        )
        if href
    )
    if not link_items:
        return ''
    return " | ".join(link_items)


class InfoTool(WaylayResource):
    """Client tool that exposes information about the provided services, recources and operations."""

    _service_context: WaylayServiceContext

    actions: Dict[str, Any] = {
        'action_info_df': {'wrapped_actions': []},
        'action_info_html': {'wrapped_actions': []}
    }

    def set_context(self, service_context):
        """Configure this tool with a service context."""
        self._service_context = service_context

    def action_info_df(
        self,
        service: Optional[Union[str, List[str]]] = None,
        resource: Optional[Union[str, List[str]]] = None,
        links: Optional[Union[str, List[str]]] = None,
        desc_levels: Optional[Union[str, List[str]]] = None,
        include_private: bool = False
    ) -> pd.DataFrame:
        """Produce a pandas DataFrame with an overview of the provided actions.

Arguments:
    service     filter on the services to be listed
    resource    filter on the resources to be listed
    links       filter on the documentation link names
        """

        def service_filter(name):
            return service is None or name == service or name in service

        def resource_filter(name):
            return resource is None or name == resource or name in resource

        def format_filtered_links(doc_links):
            if not doc_links:
                return ''
            return format_links_html(
                {
                    link: ref
                    for link, ref in doc_links.items()
                    if not links or link in links
                }
            )

        def format_pydoc(docstr: Optional[str]) -> Optional[str]:
            if not docstr:
                return docstr
            return f'<pre>{docstr.strip()}\n</pre>'

        def format_rest_action(docstr: Optional[str]) -> Optional[str]:
            if not docstr:
                return docstr
            return f'<b><pre>{docstr.strip()}\n</pre></b>'

        lvls = desc_levels or DESC_LEVELS

        def action_doc(action: WaylayAction) -> str:
            return "\n".join(f'<div>{doc}</div>' for doc in [
                action.description if DESC_DESC in lvls else None,
                format_pydoc(action.arguments_doc) if DESC_ARGS in lvls else None,
                format_pydoc(action.returns_doc) if DESC_RETURN in lvls else None,
                format_pydoc(action.sdk_action_doc) if DESC_SDK_ACTION in lvls else None,
                *([
                    wrapped_doc
                    for wrapped_action in cast(WaylayAction, getattr(action, 'wrapped_actions', []))
                    for wrapped_doc in [
                        'Can use the following REST action:',
                        format_rest_action(wrapped_action.rest_action_doc) if DESC_REST in lvls else None,
                        wrapped_action.description if DESC_DESC in lvls else None,
                        format_pydoc(wrapped_action.arguments_doc) if DESC_ARGS in lvls else None,
                        format_pydoc(wrapped_action.returns_doc) if DESC_RETURN in lvls else None,
                        format_pydoc(wrapped_action.sdk_action_doc) if DESC_SDK_ACTION in lvls else None,
                    ]
                ] if DESC_WRAPPED in lvls else []),
                format_rest_action(action.rest_action_doc) if DESC_REST in lvls else None,
                format_filtered_links(action.doc_links) if DESC_LINK in lvls else None,
            ] if doc)

        df_doc = pd.DataFrame(np.transpose([
            [
                service.service_key,
                resource.resource_name,
                action.name,
                action_doc(action),
            ]
            for service in self._service_context.list()
            if service_filter(service.name)
            for resource in service.resources
            if resource_filter(resource.name)
            for action_id, action in resource.actions.items()
            if include_private or not action.name.startswith('_')
        ]), index=[
            'service',
            'resource',
            'action',
            'description'
        ], ).T
        return df_doc.set_index(['service', 'resource', 'action'])

    def action_info_html(
        self,
        service: Optional[Union[str, List[str]]] = None,
        resource: Optional[Union[str, List[str]]] = None,
        links: Optional[Union[str, List[str]]] = None,
        desc_levels: Optional[Union[str, List[str]]] = None
    ) -> str:
        """Render the service/resource/action listing as an html table.

Arguments:
    service     filter on the services to be listed
    resource    filter on the resources to be listed
    links       filter on the documentation link names
        """
        html = self.action_info_df(service, resource, links, desc_levels).to_html(
            escape=False,
            header=False,
            index_names=False,
        )
        # custom header
        html = html.replace('<tbody>', """<thead>
  <tr>
    <th>service</th>
    <th>resource</th>
    <th>action</th>
    <th>description</th>
  </tr>
</thead>
<tbody>""")
        # replace double escapes in html
        return html.replace('\\n', "\n")



"""ETL import tooling."""
__docformat__ = "google"

from typing import (
    Any, Optional, List,
    Dict, Iterator, Union
)
from enum import Enum
from dataclasses import dataclass, asdict, replace
from contextlib import contextmanager
from pathlib import PosixPath
import logging

import pandas as pd

from waylay.exceptions import WaylayError, RestResponseError
from waylay.service import WaylayResource, WaylayAction
from waylay.service.storage import StorageService
from waylay.service.etl import ETLService
from waylay.service.resources import ResourcesService
from waylay.service.data import DataService
from waylay.service.queries import QueriesService
from waylay.service._base import WaylayServiceContext

from .parser.model import (
    SeriesInput,
    CSVOutput,
    ArchiveType,
    SeriesSettings,
    WaylayETLSeriesImport,
    ETLFile,
    PathLike,
    ETL_IMPORT_BUCKET,
    CSVReader,
    Resource,
)

from .parser import (
    prepare_etl_import,
    create_etl_import,
    dataframe_from_iterator,
    read_etl_import,
    read_etl_import_as_stream,
    list_resources,
)

from .parser import export_series

LOG = logging.getLogger(__name__)

ETL_IMPORT_TIMESERIES_FILE_SUFFIX = '-timeseries.csv.gz'
ETL_IMPORT_RESULT_FILE_SUFFIX = '.result.json'
ETL_IMPORT_SUFFIXES = [ETL_IMPORT_TIMESERIES_FILE_SUFFIX, ETL_IMPORT_RESULT_FILE_SUFFIX]


class ETLImportError(WaylayError):
    """Error raised when the ETL tool is not able to fullfill the requested operation."""


class ResourceUpdateLevel(str, Enum):
    """The possible detail levels at which Waylay Resources can be updated during an ETL import job."""

    NONE = 'none'
    ID = 'id'
    ALL = 'all'

    def __str__(self):
        """Get the string representation."""
        return self.value


class ETLImportStatus(str, Enum):
    """The possible states of an ETL import job.

    These correspond to the subfolders of the 'etl-import' bucket (except for 'not_found').
    """

    UPLOAD = 'upload'
    BUSY = 'busy'
    FAILED = 'failed'
    IGNORED = 'ignored'
    DONE = 'done'
    NOT_FOUND = 'not_found'

    def __str__(self):
        """Get the string representation."""
        return self.value


ETL_IMPORT_STATUS_MESSAGE = {
    ETLImportStatus.UPLOAD:
    'The upload has not been picked up by the ETL process. '
    'Please contact support if this persists.',
    ETLImportStatus.BUSY:
    'The upload is being processed by the ETL process.',
    ETLImportStatus.FAILED:
    'The upload could not be processed by the ETL process. Pleasy try again or contact support.',
    ETLImportStatus.IGNORED:
    'The upload did not satisfy the formatting rules of an ETL import.',
    ETLImportStatus.DONE:
    'The upload has been processed by the ETL process. Please inspect the processing report.',
    ETLImportStatus.NOT_FOUND:
    'The upload could not be found on the storage server. Please try again.'

}
ETL_IMPORT_UPLOAD_PROCESS_STATUSES = [
    ETLImportStatus.BUSY,
    ETLImportStatus.FAILED,
    ETLImportStatus.DONE
]

ETL_IMPORT_ALLOW_CLEANUP_STATUSES = [
    ETLImportStatus.UPLOAD,
    ETLImportStatus.FAILED,
    ETLImportStatus.DONE,
    ETLImportStatus.IGNORED,
]


@dataclass
class ETLImportJob:
    """Response object for the status of an etl upload.

    Reports the status of an ETL import job as executed by the ETL Server and recorded in
    the 'etl-import' storage bucket.

    Attributes:
        name: The (base) name of the ETL file as a storage object in the 'etl-import' bucket.
        message: A human readable status message.
        status: The import job status.
        storage: A link to the storage location.
        etl_import:
            The response object of the ETL Server for this job,
            as stored on the 'etl-import' bucket.
        last_etl_import:
            The response object of the last job handled by the ETL Server.
            Only included when inspecting the status of a specific upload ('check_import').
    """

    name: str
    message: str
    status: ETLImportStatus
    storage_name: Optional[str] = None
    storage_get_url: Optional[str] = None
    etl_import: Optional[Dict] = None
    last_etl_import: Optional[Dict] = None

    @property
    def storage(self) -> str:
        """Get a html link to the import file as stored on Waylay Storage."""
        if not self.storage_name:
            return ''
        return f'<a href="{self.storage_get_url}">{self.storage_name}</a>'

    @property
    def storage_folder(self) -> Optional[str]:
        """Get the storage folder in which the import file and its status report resides."""
        if not self.storage_name:
            return None
        return str(PosixPath(self.storage_name).parent)

    def to_dict(self) -> Dict[str, Any]:
        """Return a dict representation."""
        return asdict(self)

    def to_dataframe(self) -> pd.DataFrame:
        """Return a dataframe representation."""
        data_dict = asdict(self)
        data_entries = [
            (('Upload', k), format(v))
            for k, v in data_dict.items()
            if k in ['name', 'message', 'status']
            if v
        ]
        if self.storage:
            data_entries.append(
                (('Upload', 'storage'), format(self.storage))
            )
        if self.etl_import:
            data_entries.extend(
                (('ETL Import Result', k), format(v))
                for k, v in self.etl_import.items()
            )
        if self.last_etl_import:
            data_entries.extend(
                (('Last ETL Import Result', k), format(v))
                for k, v in self.last_etl_import.items()
            )
        return pd.DataFrame(
            columns=[''],
            index=pd.MultiIndex.from_tuples(
                [d[0] for d in data_entries],
                names=('', '')
            ),
            data=[format(d[1]) for d in data_entries]
        )

    def to_html(self) -> str:
        """Return an HTML string representation."""
        return self.to_dataframe().to_html(escape=False)


class TimeSeriesETLTool(WaylayResource):
    """Client tool to prepare and upload or download timeseries data.

    Attributes:
        storage_service:
            Storage service client used by this tool.
        etl_service:
            ETL service client used by this tool.
        temp_dir:
            Default temporary directory used for processes initiated by this tool.
    """

    actions: Dict[str, Any] = {
        'prepare_import': {'wrapped_actions': []},
        'initiate_import': {'wrapped_actions': []},
        'list_import':  {'wrapped_actions': []},
        'check_import':  {'wrapped_actions': []},
        'cleanup_import':  {'wrapped_actions': []},
        'read_import_as_csv':  {'wrapped_actions': []},
        'read_import_as_dataframe':  {'wrapped_actions': []},
        'list_import_resources':  {'wrapped_actions': []},
        'update_resources':  {'wrapped_actions': []},
        'export_series_as_csv':  {'wrapped_actions': []},
        'update_query':  {'wrapped_actions': []},
    }

    storage_service: StorageService
    etl_service: ETLService
    resources_service: ResourcesService
    query_service: QueriesService
    data_service: DataService
    temp_dir: Optional[PathLike] = None

    def set_context(self, service_context: WaylayServiceContext):
        """Configure this tool with a service context."""
        self.storage_service = service_context.require(StorageService)
        self.etl_service = service_context.require(ETLService)
        self.resources_service = service_context.require(ResourcesService)
        self.data_service = service_context.require(DataService)
        self.query_service = service_context.require(QueriesService)

    def prepare_import(
        self,
        *series: SeriesInput,
        name: Optional[str] = None,
        temp_dir: Optional[PathLike] = None,
        settings: Optional[SeriesSettings] = None,
        progress: bool = True,
        **settings_args: Any
    ) -> WaylayETLSeriesImport:
        """Convert an input data set to a locally stored timeseries file.

This created file can be ingested by the waylay system as timeseries data.

Arguments:
    series: A supported CSV or pandas Dataframe input source.
    name:
        A name for the upload job, will be used in naming the ETL file.
        (default: 'import-{timestamp}')
    temp_dir:
        The local storage location to use in preparing the ETL file.
        (default: a system generated temp file with prefix 'etl-import')
    settings:
        A full specification object for the data conversion. Alternatively,
        attributes of this `SeriesSettings` object can be specified as keyword
        argument.
    progress:
        Write a progress bar for the conversion to standard output.
    **settings_args:
        Any attribute of a `SeriesSettings` object. See `settings` above.
        """
        etl_file = ETLFile(temp_dir or self.temp_dir, name)
        etl_import = prepare_etl_import(*series, import_file=etl_file, settings=settings, **settings_args)
        return create_etl_import(etl_import, progress=progress)

    def initiate_import(
        self,
        etl_import: WaylayETLSeriesImport,
        resource_update_level: ResourceUpdateLevel = ResourceUpdateLevel.ID,
        progress: bool = False
    ) -> WaylayETLSeriesImport:
        """Upload a prepared timeseries file to the etl-import ingestion bucket.

Arguments:
    etl_import:
        A reference object to the input source and converted etl file, as
        produced by `prepare_import`.
        """
        self.storage_service.content.put(
            etl_import.storage_bucket,
            etl_import.storage_object_name,
            from_file=etl_import.import_file.path,
            content_type='application/gzip',
            progress=progress
        )
        if resource_update_level:
            try:
                self.update_resources(etl_import, resource_update_level)
            except Exception as e:
                LOG.warning(f"Unexpected error while trying to update resources: %s", e)
        return etl_import

    def list_import(
        self,
        name_filter: Optional[str] = None,
        status_filter: Optional[List[ETLImportStatus]] = None
    ) -> List[ETLImportJob]:
        """List all imports in all states on the `etl-import` bucket.

This queries both the content of the `etl-import` bucket,
and the current or last job held by the ETL Service.

Arguments:
    name_filter:
        A query filter on the name of the etl import files to consider.
    status_filter:
        A filter on the status of the import jobs, a list of strings
        as enumerated in `ETLImportStatus`

Returns:
    A list of 'ETLImportJob' objects that represent the processing status
    of an ETL import job.
        """
        # query storage objects in 'etl-import' bucket
        if name_filter:
            name_filter = name_filter.replace(ETL_IMPORT_TIMESERIES_FILE_SUFFIX, '')

        etl_upload_objects = [
            obj['name']
            for obj in self.storage_service.object.iter_list_all(
                ETL_IMPORT_BUCKET, '', params=dict(recursive=True, max_keys=900)
            )
            if any(
                obj['name'].startswith(f'{status}/')
                for status in ETLImportStatus
                if (not status_filter or status in status_filter)
            )
            if any(
                obj['name'].endswith(suffix)
                for suffix in ETL_IMPORT_SUFFIXES
            )
            if (name_filter is None or name_filter in obj['name'])
        ]

        response_by_key: Dict[str, ETLImportJob] = {}
        # create a ETLImportJob per timeseries upload file
        for obj_name in etl_upload_objects:
            if obj_name.endswith(ETL_IMPORT_TIMESERIES_FILE_SUFFIX):
                key = obj_name.replace(ETL_IMPORT_TIMESERIES_FILE_SUFFIX, '')
                status = obj_name.split('/')[0]
                get_url = self.storage_service.object.sign_get(ETL_IMPORT_BUCKET, obj_name)
                resp = ETLImportJob(
                    name=obj_name.split('/')[-1],
                    status=status,
                    message=ETL_IMPORT_STATUS_MESSAGE[status],
                    storage_get_url=get_url,
                    storage_name=obj_name,
                )
                response_by_key[key] = resp

        # amend a ETLImportJob for each result file
        for obj_name in etl_upload_objects:
            if obj_name.endswith(ETL_IMPORT_RESULT_FILE_SUFFIX):
                key = obj_name.replace(ETL_IMPORT_RESULT_FILE_SUFFIX, '')
                upload_file_resp = response_by_key.get(key)
                if upload_file_resp:
                    resp_content = self.storage_service.content.get(ETL_IMPORT_BUCKET, obj_name)
                    upload_file_resp.etl_import = resp_content.json()  # type: ignore

        return list(response_by_key.values())

    def check_import(
        self, etl_import: Union[str, ETLImportJob, WaylayETLSeriesImport]
    ) -> ETLImportJob:
        """Validate the status of an import process started by this tool.

Arguments:
    etl_import:
        Either a `WaylayETLSeriesImport`, representing a prepared import task,
        as produced by `initiate_import`,
        or an `ETLImportJob`, representing the status of that task,
        as returned by `list_import`, `check_import`
        or a the name of an import job.
Returns:
    An object representing the current status of the ETL import job.
        """
        if isinstance(etl_import, str):
            import_name = etl_import
        elif isinstance(etl_import, ETLImportJob):
            import_name = etl_import.name
        else:
            import_name = etl_import.import_file.name

        import_listing = [
            import_job
            for import_job in self.list_import(name_filter=import_name)
            if import_job.name == import_name
        ]

        if not import_listing:
            return ETLImportJob(
                name=import_name,
                status=ETLImportStatus.NOT_FOUND,
                message=ETL_IMPORT_STATUS_MESSAGE[ETLImportStatus.NOT_FOUND],
            )
        response = import_listing[0]

        if len(import_listing) > 1:
            listing = ''.join(f'<li>{import_job.storage}</li>' for import_job in import_listing)
            response.message = response.message + (
                '<br>Multiple import jobs with identically named import files have been found:'
                f'<ul>{listing}</ul>'
                'Please provide unique names for each import file.'
            )

        # current etl import task
        try:
            current_etl_import = self.etl_service.etl_import.get()
        except RestResponseError as exc:
            action: WaylayAction = self.etl_service.etl_import.get.action
            current_etl_import = dict(
                message=(
                    'Status of last import not available at '
                    f'`{action.rest_action_doc}`'
                ),
                cause=exc.message
            )

        # only add if different from saved etl import result for current job
        if (
            not response.etl_import
            or response.etl_import.get('filename', '') != current_etl_import.get('filename')
        ):
            response.last_etl_import = current_etl_import

        # warn if current etl is not the same as selected
        if 'filename' in current_etl_import:
            current_etl_filename = current_etl_import['filename']
            if import_name not in current_etl_filename:
                response.message = response.message + \
                    '<br>The ETL service is/has currently been processing another import.'

        return response

    def cleanup_import(
        self, etl_import: Union[str, WaylayETLSeriesImport, ETLImportJob], force: bool = False
    ) -> ETLImportJob:
        """Clean up the server storage for an import task.

Arguments:
    etl_import:
        Either a `WaylayETLSeriesImport`, representing a prepared import task,
        as produced by `initiate_import`,
        or an `ETLImportJob`, representing the status of that task,
        as returned by `list_import`, `check_import`.
    force:
        If true, delete storage even if the import job is reported
        to be running or not existing.

Returns:
    An object representing the current status of the ETL import job.

        """
        etl_import_job = self.check_import(etl_import)
        storage_folder = etl_import_job.storage_folder

        if not storage_folder:
            LOG.warning('No matching import job found for %s.', etl_import_job)
            return etl_import_job

        if not force and etl_import_job.status not in ETL_IMPORT_ALLOW_CLEANUP_STATUSES:
            raise ETLImportError(
                f'Cannot cleanup import job `{etl_import_job.name}`, '
                f'as its status is `{etl_import_job.status}``. '
                'Use `force=True` argument to override this status check.'
            )

        resp = self.storage_service.folder.remove(
            ETL_IMPORT_BUCKET, storage_folder, params={'recursive': 'true'}
        )
        for link in resp.get('removed', []):
            LOG.info('removed: %s', link.get("href", link))
        return self.check_import(etl_import_job)

    @contextmanager
    def read_import_as_csv(
        self, etl_import: WaylayETLSeriesImport
    ) -> Iterator[CSVReader]:
        """Read an etl import file as a csv stream.

This returns a context manager that provides an iterator
of csv rows (`Iterator[Sequence[str]]`):

    with etl_tool.read_import_as_csv(etl_import) as csv_data:
        for csv_line in csv_data:
            print("|".join(csv_line))

Arguments:
    etl_import:
        The object representing the converted ETL import file,
        as produced by `prepare_import`.

Returns:
    A context manager providing an iterator of string records.

        """
        with read_etl_import_as_stream(etl_import) as csv_iterator:
            yield csv_iterator

    def read_import_as_dataframe(self, etl_import: WaylayETLSeriesImport) -> pd.DataFrame:
        """Read an etl import file as pandas dataframe.

Arguments:
    etl_import:
        The object representing the converted ETL import file,
        as produced by `prepare_import`.

Returns:
    A pandas Dataframe containing the import data.
    Each series is a seperate column. Column headers
    contain a `resource` and `metric` reference.
    When the input specification of the `etl_import` contains
    a `metrics` object with a `value_type` or `value_parser` attribute,
    the corresponding csv series is converted to that data type.
        """
        with read_etl_import(etl_import) as series_iterator:
            return dataframe_from_iterator(series_iterator)

    def list_import_resources(
        self,
        etl_import: WaylayETLSeriesImport
    ) -> List[Resource]:
        """List resource and metric metadata contained in an ETL import file.

Arguments:
    etl_import:
        The object representing the converted ETL import file,
        as produced by `prepare_import`.

Returns:
    A list of `Resource` metadata objects describing the
    content of the import file.
        """
        with read_etl_import(etl_import) as series_iterator:
            return list_resources(etl_import.settings, series_iterator)

    def update_resources(
        self,
        etl_import: WaylayETLSeriesImport,
        resource_update_level: ResourceUpdateLevel = ResourceUpdateLevel.ALL
    ) -> List[Resource]:
        """Create or update the Waylay Resources for the timeseries in this dataset.

Arguments:
    etl_import:
        The object representing the converted ETL import file,
        as produced by `prepare_import`.

Returns:
    A list of `Resource` metadata objects describing the
    content of the import file, and for which metadata has been updated
    to the waylay system
        """
        resources = self.list_import_resources(etl_import)
        for resource in resources:
            if resource_update_level == ResourceUpdateLevel.ID:
                self.resources_service.resource.update(resource.id, body={})

            if resource_update_level == ResourceUpdateLevel.ALL:
                self.resources_service.resource.update(resource.id, body=resource.to_dict())
        return resources

    def update_query(
        self,
        etl_import: WaylayETLSeriesImport,
        name: Optional[str] = None,
        **query_params: Any
    ) -> Any:
        """Create or update a waylay query containing all the series defined in this import.

The name of the import will be used as query name.
Arguments:
    etl_import:
        The object representing the converted ETL import file,
        as produced by `prepare_import`.
    name:
        The name of the query. Defaults to the prefix name of
        the etl_import argument (`etl_import.import_file.prefix`).
    query_params:
        Any additional query params (like `from`, `util`, `freq`, `aggregation`, ...)
        to be used in the query

Returns:
    The result of a `queries.query.replace()` SDK call that
    updated or created the query with the name of the import.
        """
        query_name = name or etl_import.import_file.prefix
        resources = self.list_import_resources(etl_import)
        with self.read_import_as_csv(etl_import) as csv_lines:
            header = next(csv_lines)
            ts_idx = header.index(etl_import.settings.timestamp_column)
            first_timestamp = next(csv_lines)[ts_idx]
        query = {
            'from': first_timestamp,
            'data': [
                {
                    'resource': resource.id,
                    'metric': metric.name
                }
                for resource in resources
                for metric in resource.metrics or []
            ]
        }
        query.update(query_params)
        return self.query_service.query.replace(query_name, body={
            'query': query,
            'meta': {
                'description': 'Generated by `timeseries.etl_tool.update_query` (python SDK), '
                f'for etl_import via {etl_import.storage_object_name}.'
            }
        })

    def export_series_as_csv(
        self,
        output: CSVOutput,
        archive_type: Optional[str] = None,
        progress: bool = True,
        settings: Optional[SeriesSettings] = None,
        timestamp_column: Optional[str] = 'timestamp',
        **settings_args: Any,
    ):
        """Export a number of series from the times series database.

    Arguments:
        output:
            The object to which the csv lines will be written. Can be a file path, stream, etc.
        archive_type:
            Optional archive type.
        progress:
            Optional boolean to write a progress bar for the export.
        settings:
            Optional series settings.
        timestamp_column:
            Optional name for timestamp column. Defaults to 'timestamp'. Set to None if not wanted.
        """
        export_settings = settings or SeriesSettings()
        export_settings = replace(export_settings, **settings_args)
        # set timestamp column (default `timestamp`) unless explicitely set to None
        if timestamp_column:
            export_settings.timestamp_column = timestamp_column
        series_provider = export_series.PagingExportSeriesProvider(export_settings, self.data_service.series)
        return export_series.export_csv(
            output,
            export_settings,
            series_provider,
            ArchiveType.lookup(archive_type),
            progress,
        )



"""Conversion from CSV to elt export format."""

from collections import defaultdict
from enum import Enum
from typing import (
    Union, Dict, Sequence, Iterator, Iterable,
    Optional, Tuple, List, TextIO,
    cast,
)
from dataclasses import replace
import collections.abc
import os
import io
import csv
import gzip
import itertools
import tarfile
import pathlib
import logging
from contextlib import contextmanager
from datetime import datetime
from tarfile import TarFile
from zipfile import ZipFile

from tqdm import tqdm
import pandas as pd
from .model import (
    ArchiveType,
    SeriesInput,
    SeriesSettings,
    Measurement,
    MeasurementIterator,
    Metric,
    CSVReader,
    CSVReaderAndResource,
    PathLike,
    ETLFile,
    ETL_IMPORT_COLUMN_NAMES,
    RESOURCE_COLUMN,
    METRIC_COLUMN,
    VALUE_COLUMN,
    ParserRequestError,
    TIMESTAMP_COLUMN_NAMES,
)

from .util import (
    WrappeableSpooledTemporaryFile,
)

LOG = logging.getLogger(__name__)
TIMESERIES_SUFFIXES = ['timeseries.csv.gz', '-timeseries.csv']
DEFAULT_SPOOL_MAX_SIZE = 2 << 18

_SPOOLED_TS_LEN = 13
_SPOOLED_TS_FORMAT = '%013d'


class MeasurementsStoreType(Enum):
    """Enum for measurements store types."""

    MEMORY = 'memory'
    SPOOLED = 'spooled'


class CsvImportSeriesProvider(collections.abc.Mapping):
    """A series provider that uses cached buffers to process import csv files."""

    def __init__(
        self,
        settings: SeriesSettings,
        measurements_store: Union[str, MeasurementsStoreType] = MeasurementsStoreType.SPOOLED
    ):
        """Create an import series provider."""
        self.settings = settings
        if isinstance(measurements_store, MeasurementsStoreType):
            self._measurements_store = measurements_store.value
        elif measurements_store in [mt.value for mt in MeasurementsStoreType]:
            self._measurements_store = measurements_store
        else:
            logging.warning(
                "Series are sorted using limited memory buffers with overflow to file. "
                "This might use up the open file quota when importing many different series. "
                "Use measurements_store='memory' for a pure in-memory based solution. "
                "Use measurements_store='spooled' to suppress this warning"
            )
            self._measurements_store = MeasurementsStoreType.SPOOLED.value
        self._measurements: Dict[Tuple[str, str], Union[TextIO, io.StringIO]] = defaultdict(self._new_buffer)
        self._reading = False
        self._written = 0
        self.row_count = 0
        self.input_count = 0
        self.spool_size_left = DEFAULT_SPOOL_MAX_SIZE

    def _new_buffer(self) -> Union[TextIO, io.StringIO]:
        if self._measurements_store == MeasurementsStoreType.MEMORY.value:
            return io.StringIO()
        else:
            max_size = int(self.spool_size_left / 2)
            self.spool_size_left -= max_size
            return io.TextIOWrapper(WrappeableSpooledTemporaryFile(max_size=max_size, mode='w+b'))

    def __enter__(self):
        """Enter a runtime context."""
        assert not self._reading
        return self

    def __exit__(self, exc, value, tb):
        """Leave a runtime context, cleaning up all buffers."""
        self.close()

    def close(self):
        """Close all buffers."""
        for buffer in self._measurements.values():
            buffer.close()

    def __iter__(self):
        """Iterate over the (resource, metric) keys of this provider."""
        return iter(self._measurements)

    def __len__(self):
        """Get the number of requested (resource,metric) combinations."""
        return len(self._measurements)

    def __getitem__(self, resource_metric: Tuple[str, str]) -> Iterable[Measurement]:
        """Create an measurement iterable for a given resource, metric."""
        resource, metric = resource_metric
        provider = self

        class _MeasurementIterable():
            def __iter__(self) -> Iterator[Measurement]:
                return provider._iter_single_series_export(resource, metric)

        return _MeasurementIterable()

    @staticmethod
    def _write_measurement_to_buffer(target: Union[io.StringIO, TextIO], timestamp: datetime, value):
        target.write(_SPOOLED_TS_FORMAT % int(timestamp.timestamp() * 1000))
        target.write(str(value))
        target.write('\n')

    @staticmethod
    def _peek_reader(csv_data: CSVReader) -> Tuple[Sequence[str], Optional[Sequence[str]], CSVReader]:
        header = next(csv_data)
        first_data_line: Optional[Sequence[str]] = []
        while first_data_line == []:
            first_data_line = next(csv_data, None)

        # restore the iterator from which we read the first line
        if first_data_line:
            csv_data = itertools.chain([first_data_line], csv_data)
        return header, first_data_line, csv_data

    def _prepare_settings(
        self, header: Sequence[str],
        example_data: Sequence[str],
        default_resource_id: Optional[str]
    ) -> SeriesSettings:
        settings = _prepare_csvheader_timestamp(self.settings, header, example_data)

        if default_resource_id and settings.resource is None:
            settings = replace(settings, resource=default_resource_id)
        settings = _prepare_csvheader_resource_metric(header, settings)
        if settings.value_column or all(column in header for column in ETL_IMPORT_COLUMN_NAMES):
            # single column contains values, metric must be provided fixed or as other column.
            settings = _prepare_csvheader_value_column(header, settings)
        else:
            # series specified in seperate columns
            settings = _prepare_csvheader_series_column(header, settings)
        return settings

    @staticmethod
    def _header_indices(
        header: Sequence[str],
        settings: SeriesSettings
    ):
        timestamp_idx = header.index(settings.timestamp_column)
        resource_idx = None
        metric_idx = None
        value_idx = None
        if settings.resource_column is not None:
            resource_idx = header.index(settings.resource_column)
        if settings.metric_column is not None:
            metric_idx = header.index(settings.metric_column)
        if settings.value_column is not None:
            value_idx = header.index(settings.value_column)

        return resource_idx, metric_idx, timestamp_idx, value_idx

    def add_csv_data(self, csv_data: CSVReader, default_resource_id: Optional[str]) -> int:
        """Add a csv data to the measurement buffer."""
        if self._reading:
            raise RuntimeError('Cannot write to import buffer after reading from it.')

        self.input_count += 1

        header, first_data_line, csv_data = self._peek_reader(csv_data)
        if first_data_line is None:
            # empty csv file
            return 0

        settings = self._prepare_settings(header, first_data_line, default_resource_id)

        create_timestamp = settings.get_timestamp_parser()
        assert create_timestamp

        resource_idx, metric_idx, timestamp_idx, value_idx = self._header_indices(header, settings)

        if value_idx is not None:
            # normalised, single-value csv
            def _handle_csv_normalized(item):
                resource_id = settings.resource_by_key(item[resource_idx] if resource_idx is not None else None)
                metric_name = settings.metric_by_key(item[metric_idx] if metric_idx is not None else None)
                if resource_id and metric_name:
                    self._write_measurement_to_buffer(
                        self._measurements[(resource_id, metric_name)],
                        create_timestamp(item[timestamp_idx]),
                        item[value_idx]
                    )
                    return 1
                return 0

            item_handler = _handle_csv_normalized
        elif resource_idx is not None:
            # value in columns with metric key, with resource column
            value_idx_and_metric_name: List[Tuple[int, str]] = []
            for value_idx, metric_key in enumerate(header):
                if value_idx in (resource_idx, timestamp_idx):
                    continue
                metric_name = settings.metric_by_key(metric_key)
                if metric_name is not None:
                    value_idx_and_metric_name.append((value_idx, metric_name))
            value_count = len(value_idx_and_metric_name)

            def _handle_csv_with_resource(item):
                resource_id = settings.resource_by_key(item[resource_idx])
                if resource_id is None:
                    return 0
                for value_idx, metric_name in value_idx_and_metric_name:
                    self._write_measurement_to_buffer(
                        self._measurements[(resource_id, metric_name)],
                        create_timestamp(item[timestamp_idx]),
                        item[value_idx]
                    )
                return value_count

            item_handler = _handle_csv_with_resource
        else:
            # value in columns with metric key, no resource column
            resource_id = settings.resource
            if resource_id is None:
                raise ParserRequestError('No default resource id provided.')

            value_idx_and_measurements: List[Tuple[int, TextIO]] = []
            for value_idx, metric_key in enumerate(header):
                if value_idx in (resource_idx, timestamp_idx):
                    continue
                metric_name = settings.metric_by_key(metric_key)
                if metric_name:
                    measurements = self._measurements[(resource_id, metric_name)]
                    value_idx_and_measurements.append((value_idx, measurements))
            value_count = len(value_idx_and_measurements)

            def _handle_csv_without_resource(item):
                for value_idx, measurements in value_idx_and_measurements:
                    self._write_measurement_to_buffer(
                        measurements,
                        create_timestamp(item[timestamp_idx]),
                        item[value_idx]
                    )
                return value_count
            item_handler = _handle_csv_without_resource

        row_count = 0
        for item in csv_data:
            # skip empty lines
            if item:
                row_count += item_handler(item)
        self.row_count += row_count
        return row_count

    def flip(self):
        """Switch from writing to reading mode."""
        if self._reading:
            return
        self._reading = True
        for buffer in self._measurements.values():
            self._written += buffer.tell()
            buffer.seek(0)

    def _iter_single_series_export(
        self,
        resource_id: str, metric_name: str
    ) -> MeasurementIterator:
        """Paging iterator over the specified resource and metric."""
        self.flip()
        measure_buffer = self._measurements[(resource_id, metric_name)]
        measure_buffer.seek(0)
        for line in measure_buffer:
            yield (
                pd.Timestamp(int(line[:_SPOOLED_TS_LEN]), unit='ms', tz='UTC'),
                # do not read \n
                line[_SPOOLED_TS_LEN:-1]
            )


def prepare_settings_csv(
    *input_data: SeriesInput,
    settings: SeriesSettings,
) -> SeriesSettings:
    """Validate and update the input settings by extracting metadata from a csv header."""
    original_settings = settings
    input_count = 0
    for csv_data, default_resource_id in open_csv(*input_data):
        input_count += 1
        header = next(csv_data)
        example = next(csv_data, None)
        settings = original_settings
        if not settings.resource:
            settings = replace(settings, resource=default_resource_id)

        settings = _prepare_csvheader_timestamp(settings, header, example)

        settings = _prepare_csvheader_resource_metric(header, settings)

        if settings.value_column or all(column in header for column in ETL_IMPORT_COLUMN_NAMES):
            # single column contains values, metric must be provided fixed or as other column.
            settings = _prepare_csvheader_value_column(header, settings)
        else:
            # series specified in seperate columns
            settings = _prepare_csvheader_series_column(header, settings)

    if input_count == 0:
        raise ParserRequestError('No input data')
    if input_count == 1:
        return settings
    return original_settings


def _prepare_csvheader_timestamp(
    settings: SeriesSettings, header: Sequence[str], example: Optional[Sequence[str]]
) -> SeriesSettings:
    # time column handling
    if settings.timestamp_column:
        if settings.timestamp_column not in header:
            raise ParserRequestError(
                f'Timestamp column `{settings.timestamp_column}` not in csv header {header}.'
            )
    elif settings.timestamp_first and settings.timestamp_interval:
        raise NotImplementedError('timestamp_first timestamp_interval setting not supported')
    else:
        for timestamp_column in TIMESTAMP_COLUMN_NAMES:
            if timestamp_column in header:
                settings = replace(settings, timestamp_column=timestamp_column)
                break
        if not settings.timestamp_column:
            raise ParserRequestError(
                f'No timestamp column found in `{header}`'
            )

    # default timestamp value constructor
    if settings.timestamp_column:
        timestamp_idx = next(
            idx for idx, col in enumerate(header)
            if col == settings.timestamp_column
        )
        example_timestamp = example[timestamp_idx] if example else None
        timestamp_constructor = settings.get_timestamp_parser(example_timestamp)
        if settings.timestamp_offset:
            offset = settings.timestamp_offset
            original_timestamp_constructor = timestamp_constructor

            def timestamp_constructor_offset(ts):
                return original_timestamp_constructor(ts) + offset

            timestamp_constructor = timestamp_constructor_offset

        settings = replace(
            settings,
            timestamp_constructor=timestamp_constructor
        )
    return settings


def _prepare_csvheader_resource_metric(header: Sequence[str], settings: SeriesSettings) -> SeriesSettings:
    if not settings.resource_column:
        if RESOURCE_COLUMN in header:
            settings = replace(settings, resource_column=RESOURCE_COLUMN)
        elif settings.resource is None:
            raise ParserRequestError(
                f'No valid resource column or default provided: `{settings.resource_column}`'
            )
    if settings.resource_column:
        if settings.resource_column not in header:
            raise ParserRequestError(
                f'Invalid resource column provided: `{settings.resource_column}`'
            )
    if settings.metric_column:
        if settings.metric_column not in header:
            raise ParserRequestError(
                f'Invalid metric column provided: `{settings.metric_column}`'
            )
    return settings


def _prepare_csvheader_value_column(header: Sequence[str], settings: SeriesSettings) -> SeriesSettings:
    if not settings.value_column:
        settings = replace(settings, value_column=VALUE_COLUMN)
    if settings.value_column not in header:
        raise ParserRequestError(
            f'Value column `{settings.value_column}` not in csv header {header}.'
        )
    if settings.metric_column is None:
        if METRIC_COLUMN in header:
            settings = replace(settings, metric_column=METRIC_COLUMN)
        elif settings.metric is None:
            raise ParserRequestError(
                f'No valid metric column or default provided for values in `{settings.value_column}` column.'
            )
    return settings


def _prepare_csvheader_series_column(header: Sequence[str], settings: SeriesSettings) -> SeriesSettings:
    specified_metrics = list(settings.iter_metrics())
    if not specified_metrics:
        specified_metrics = [
            Metric(name=column)
            for column in header
            if column != settings.timestamp_column
            if column != settings.resource_column
        ]

    specified_metrics = [
        metric_spec
        for metric_spec in specified_metrics
        if metric_spec.key_or_name in header
    ]
    if not specified_metrics:
        raise ParserRequestError(
            f'None of the specified metrics `{settings.metrics}` found in the csv header.'
        )
    return replace(settings, metrics=specified_metrics)


@contextmanager
def create_import_series_provider(
    *series_input: SeriesInput,
    settings: SeriesSettings,
    progress: bool = False,
    measurements_store: MeasurementsStoreType = MeasurementsStoreType.SPOOLED
) -> Iterator[CsvImportSeriesProvider]:
    """Create a SeriesProvider from the given csv input and settings.

    This reads the csv input into buffers, ready to export to an etl file.
    Returns a context manager, that closes all temporary buffers on leaving the context.
    """
    with CsvImportSeriesProvider(settings, measurements_store) as series_provider:
        if progress:
            with tqdm(
                total=series_provider.input_count, unit='csv_files', unit_scale=True, unit_divisor=1
            ) as progress_tqdm:
                for csv_reader, default_resource_id in open_csv(*series_input):
                    series_provider.add_csv_data(csv_reader, default_resource_id)
                    progress_tqdm.update()
        else:
            for csv_reader, default_resource_id in open_csv(*series_input):
                series_provider.add_csv_data(csv_reader, default_resource_id)

        yield series_provider


def open_csv(*series_inputs: SeriesInput, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    """Open the CSV file specified by this input."""
    for series_input in series_inputs:
        yield from _open_csv(series_input, **csv_format_args)


def _open_csv(series_input: SeriesInput, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    if isinstance(series_input, ETLFile):
        LOG.debug('reading csv from %s', series_input.path)
        yield from _open_csv_gz_file(series_input.path, **csv_format_args)
    elif isinstance(series_input, (str, os.PathLike)):
        series_path = pathlib.Path(series_input)
        archive_type = ArchiveType.for_import(series_path)
        LOG.debug('reading from %s as %s', series_path, archive_type)

        if archive_type == ArchiveType.GZ:
            yield from _open_csv_gz_file(series_path, **csv_format_args)
        elif archive_type == ArchiveType.ZIP:
            yield from _open_csv_from_zip(series_path, **csv_format_args)
        elif archive_type in (ArchiveType.TAR, ArchiveType.TAR_GZ):
            yield from _open_csv_from_tar(series_path, **csv_format_args)
        elif archive_type == ArchiveType.TEXT:
            yield from _open_csv_file(series_path, **csv_format_args)
        elif archive_type == ArchiveType.DIR:
            for child_path in series_path.iterdir():
                yield from _open_csv(child_path, **csv_format_args)
        else:
            raise Exception(f'Unsupported input file {series_path} of type {archive_type}')

    elif isinstance(series_input, io.TextIOBase):
        LOG.debug('reading from text stream')
        yield from _reset_and_open_csv_text_stream(series_input, **csv_format_args)

    elif isinstance(series_input, Iterable):
        LOG.debug('reading from iterable')
        yield from _open_csv_iterable(series_input, **csv_format_args)

    else:
        raise Exception(f'Unsupported input type {type(series_input)}.')


def _open_csv_gz_file(series_input: PathLike, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    with gzip.open(series_input, 'rt') as csv_file:
        yield csv.reader(cast(io.TextIOBase, csv_file), **csv_format_args), pathlib.Path(series_input).stem


def _open_csv_file(series_input: PathLike, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    with open(series_input, 'rt') as csv_file:
        yield csv.reader(csv_file, **csv_format_args), pathlib.Path(series_input).stem


def _reset_and_open_csv_text_stream(series_input: io.TextIOBase, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    series_input.seek(0)
    yield csv.reader(series_input, **csv_format_args), None


def _open_csv_from_tar(series_input: pathlib.Path, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    with tarfile.open(series_input, 'r') as tar_archive:
        for entry_name in tar_archive.getnames():
            LOG.debug('reading tar entry %s from %s', entry_name, series_input)
            if any(entry_name.endswith(suffix) for suffix in TIMESERIES_SUFFIXES):
                yield from _open_csv_gz_from_tar_entry(tar_archive, entry_name, **csv_format_args)
            else:
                yield from _open_csv_from_tar_entry(tar_archive, entry_name, **csv_format_args)


def _open_csv_from_tar_entry(archive: TarFile, entry: str, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    csv_file = archive.extractfile(entry)
    if csv_file:
        yield csv.reader(io.TextIOWrapper(csv_file), **csv_format_args), pathlib.Path(entry).stem


def _open_csv_gz_from_tar_entry(archive: TarFile, entry: str, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    csv_gz_file = archive.extractfile(entry)
    if csv_gz_file:
        with gzip.open(csv_gz_file, 'rt') as csv_file:  # type: ignore
            yield csv.reader(cast(io.TextIOBase, csv_file), **csv_format_args), pathlib.Path(entry).stem


def _open_csv_from_zip(series_input: pathlib.Path, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    with ZipFile(series_input, 'r') as zip_archive:
        for entry_name in zip_archive.namelist():
            LOG.debug('reading zip entry %s from %s', entry_name, series_input)
            if any(entry_name.endswith(suffix) for suffix in TIMESERIES_SUFFIXES):
                yield from _open_csv_gz_from_zip_entry(zip_archive, entry_name, **csv_format_args)
            else:
                yield from _open_csv_from_zip_entry(zip_archive, entry_name, **csv_format_args)


def _open_csv_from_zip_entry(archive: ZipFile, entry: str, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    with archive.open(entry, 'r') as csv_zipped_file:
        yield csv.reader(io.TextIOWrapper(csv_zipped_file), **csv_format_args), pathlib.Path(entry).stem


def _open_csv_gz_from_zip_entry(archive: ZipFile, entry: str, **csv_format_args) -> Iterator[CSVReaderAndResource]:
    with archive.open(entry, 'r') as csv_zipped_file:
        with gzip.open(csv_zipped_file, 'rt') as csv_file:
            yield csv.reader(cast(io.TextIOBase, csv_file), **csv_format_args), pathlib.Path(entry).stem


def _open_csv_iterable(
    series_input: Union[Iterable[str], Iterable[Sequence[str]]], **csv_format_args
) -> Iterator[CSVReaderAndResource]:
    # inspect first item of iterable
    first_line = next(iter(series_input))
    if isinstance(first_line, str):
        # Iterable of strings, treat as unparsed csv data
        yield csv.reader(
            iter(cast(Iterable[str], series_input)),
            **csv_format_args
        ), None
    elif isinstance(first_line, Sequence):
        # Iterable of lists, treat as parsed csv data
        yield iter(cast(Iterable[Sequence[str]], series_input)), None
    else:
        raise ParserRequestError(
            f'Cannot read first line of a csv input:\n{first_line}'
        )
    return



"""Waylay Python SDK."""

from .client import WaylayClient
from .config import WaylayConfig
from .auth import CredentialsType, ClientCredentials
from .exceptions import WaylayError, RestResponseError
from . import _version
__version__ = _version.get_versions()['version']



"""REST definitions for the import process entity of the etl service."""

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]


class ImportResource(WaylayResource):
    """REST Resource for the import process entity of the etl service."""

    link_roots = {
        'doc': '${doc_url}/features/etl/?id='
    }

    actions = {
        'get': {
            'method': 'GET',
            'url': '/etl/import',
            'returns': [{
                'name': 'body',
                'type': 'dict',
                'description': 'A representation of the active or last ETL import process (if any).'
            }],
            'decorators': DEFAULT_DECORATORS,
            'description': "Retrieves the last or active etl import process.",
            'links': {
                'doc': 'etl-import-service',
            }
        }
    }



"""Utilities to handle waylay authentication."""

from datetime import datetime
from enum import Enum
from dataclasses import dataclass
from typing import (
    Optional, Generator, ClassVar,
    Dict, Any, List, Callable,
)
import json
import abc

import base64
import binascii
import httpx

import jwt
from jwt.exceptions import InvalidTokenError, ExpiredSignatureError, PyJWTError

from .exceptions import AuthError

# http client dependencies
_http = httpx


class CredentialsType(str, Enum):
    """Supported Waylay Authentication Methods.

    Note that username/password authentication (as used in our IDP at https://login.waylay.io)
    is not (yet) supported.
    """

    CLIENT = 'client_credentials'
    APPLICATION = 'application_credentials'
    TOKEN = 'token'
    CALLBACK = 'interactive'

    def __str__(self):
        """Get the string representation."""
        return self.value


DEFAULT_GATEWAY_URL = 'https://api.waylay.io'
DEFAULT_ACCOUNTS_URL = 'https://accounts-api.waylay.io'

ACCOUNTS_USERS_ME_PATH = '/accounts/v1/users/me'

TokenString = str


class WaylayCredentials(abc.ABC):
    """Base class for the representation of credentials to the waylay platform."""

    gateway_url: Optional[str] = None
    credentials_type: ClassVar[CredentialsType] = CredentialsType.CALLBACK

    # legacy
    accounts_url: Optional[str] = None

    @abc.abstractmethod
    def to_dict(self, obfuscate=True) -> Dict[str, Any]:
        """Convert the credentials to a json-serialisable representation."""

    @abc.abstractproperty
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        return None

    def __repr__(self):
        """Show the implementing class and public information."""
        return f'<{self.__class__.__name__}({str(self)})>'

    def __str__(self):
        """Show the credential attributes, with secrets obfuscated."""
        return json.dumps(self.to_dict(obfuscate=True))

    @abc.abstractmethod
    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed.

        This does not assure that they will lead to a succesfull authentication.
        """


CredentialsCallback = Callable[[Optional[str]], WaylayCredentials]


@dataclass(repr=False)
class AccountsUrlMixin:
    """Dataclass mixin for the 'gateway_url' (legacy 'accounts_url') property."""

    gateway_url: Optional[str] = None
    accounts_url: Optional[str] = None


@dataclass(repr=False, init=False)
class ApiKeySecretMixin(AccountsUrlMixin):
    """Dataclass mixin for the 'api_key' and 'api_secret'."""

    api_key: str = ''
    api_secret: str = ''

    def __init__(
        self, api_key: str, api_secret: str,
        *, gateway_url: Optional[str] = None, accounts_url: Optional[str] = None
    ):
        """Initialise with the api_key and api_secret."""
        super().__init__(gateway_url=gateway_url, accounts_url=accounts_url)
        self.api_key = api_key
        self.api_secret = api_secret

    @property
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        return self.api_key

    @classmethod
    def create(
        cls, api_key: str, api_secret: str,
        *, gateway_url: Optional[str] = None, accounts_url: Optional[str] = None
    ):
        """Create a client credentials object."""
        return cls(
            api_key=api_key, api_secret=api_secret, gateway_url=gateway_url, accounts_url=accounts_url
        )

    def to_dict(self, obfuscate=True):
        """Convert the credentials to a json-serialisable representation."""
        return dict(
            type=self.credentials_type.value,
            api_key=self.api_key,
            api_secret='********' if obfuscate else self.api_secret,
            gateway_url=self.gateway_url,
            accounts_url=self.accounts_url
        )

    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed.

        This does not assure that they will lead to a succesfull authentication.
        """
        if not (self.api_key and self.api_secret):
            return False
        # api key are 12 bytes hexc encoded
        try:
            if len(bytes.fromhex(self.api_key)) != 12:
                return False
        except ValueError:
            return False
        # api secret are 24 bytes base64 encoded (rfc4648)
        try:
            if len(base64.b64decode(self.api_secret, validate=True)) != 24:
                return False
        except binascii.Error:
            return False
        return True


@dataclass(repr=False, init=False)
class NoCredentials(AccountsUrlMixin, WaylayCredentials):
    """Represents that credentials can be asked via (interactive) callback when required."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.CALLBACK

    def to_dict(self, obfuscate=True):  # pylint: disable=unused-argument
        """Convert the credentials to a json-serialisable representation."""
        return dict(
            type=str(self.credentials_type),
            gateway_url=self.gateway_url,
            accounts_url=self.accounts_url
        )

    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed."""
        return True

    @property
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        return None


@dataclass(repr=False, init=False)
class ClientCredentials(ApiKeySecretMixin, WaylayCredentials):
    """Waylay Credentials: api key and secret of type 'client_credentials'."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.CLIENT


@dataclass(repr=False, init=False)
class ApplicationCredentials(ApiKeySecretMixin, WaylayCredentials):
    """Waylay Credentials: api key and secret of type 'application_credentials'."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.APPLICATION
    tenant_id: str = ''


@dataclass(repr=False, init=False)
class TokenCredentials(AccountsUrlMixin, WaylayCredentials):
    """Waylay JWT Token credentials."""

    credentials_type: ClassVar[CredentialsType] = CredentialsType.TOKEN
    token: TokenString = ''

    def __init__(
        self, token: TokenString,
        *, gateway_url: Optional[str] = None, accounts_url: Optional[str] = None
    ):
        """Create a TokenCredentials from a token string."""
        super().__init__(gateway_url=gateway_url, accounts_url=accounts_url)
        self.token = token

    @property
    def id(self) -> Optional[str]:
        """Get the main identifier for this credential."""
        try:
            token = WaylayToken(self.token)
            return f'[{token.domain}] {token.subject}'
        except AuthError as exc:
            return f'INVALID_TOKEN({exc})'

    def to_dict(self, obfuscate=True):
        """Get the credential attributes."""
        return dict(
            type=str(self.credentials_type),
            token='*********' if obfuscate else self.token,
            gateway_url=self.gateway_url,
            accounts_url=self.accounts_url
        )

    def is_well_formed(self) -> bool:
        """Validate that these credentials are well-formed."""
        try:
            # WaylayToken constructor decodes the data without signature verification
            return WaylayToken(self.token).tenant is not None
        except AuthError:
            return False


class WaylayToken:
    """Holds a Waylay JWT token."""

    def __init__(self, token_string: str, token_data: Optional[Dict] = None):
        """Create a Waylay Token holder object from given token string or data."""
        self.token_string = token_string
        if token_data is None:
            try:
                token_data = jwt.decode(token_string, options=dict(verify_signature=False))
            except (TypeError, ValueError, PyJWTError) as exc:
                raise AuthError(_auth_message_for_exception(exc)) from exc
        self.token_data = token_data

    def validate(self) -> 'WaylayToken':
        """Verify essential assertions, and its expiry state.

        This implementation does not verify the signature of a token,
        as this is seen the responsability of a server implementation.
        """
        if not self.token_string:
            raise AuthError('no token')

        if not self.token_data:
            raise AuthError('could not parse token data')

        if not self.tenant:
            raise AuthError('invalid token')

        # assert expiry
        if self.is_expired:
            raise AuthError('token expired')
        return self

    @property
    def tenant(self) -> Optional[str]:
        """Get the tenant id asserted by the token."""
        return self.token_data.get('tenant', None)

    @property
    def domain(self) -> Optional[str]:
        """Get the waylay domain asserted by the token."""
        return self.token_data.get('domain', None)

    @property
    def subject(self) -> Optional[str]:
        """Get the subject asserted by the token."""
        return self.token_data.get('sub', None)

    @property
    def licenses(self) -> List[str]:
        """Get the licenses asserted by the token."""
        return self.token_data.get('licenses', [])

    @property
    def groups(self) -> List[str]:
        """Get the groups asserted by the token."""
        return self.token_data.get('groups', [])

    @property
    def permissions(self) -> List[str]:
        """Get the permissions asserted by the token."""
        return self.token_data.get('permissions', [])

    @property
    def expires_at(self) -> Optional[datetime]:
        """Get the token expiry timestamp."""
        exp = self.token_data.get('exp', None)
        return None if exp is None else datetime.fromtimestamp(exp)

    @property
    def issued_at(self) -> Optional[datetime]:
        """Get the token issuance timestamp."""
        iat = self.token_data.get('iat', None)
        return None if iat is None else datetime.fromtimestamp(iat)

    @property
    def expires_seconds(self) -> int:
        """Get seconds until expiry."""
        exp = self.token_data.get('exp', None)
        return 0 if exp is None else exp - datetime.now().timestamp()

    @property
    def age(self) -> int:
        """Get seconds sinds issuance."""
        iat = self.token_data.get('iat', 0)
        return int(datetime.now().timestamp() - iat)

    @property
    def is_expired(self) -> bool:
        """Get the expiration state.

        True if a (previously valid) the token has expired.
        """
        if not isinstance(self.token_data, dict):
            return True
        exp = self.expires_at
        return exp is None or exp < datetime.now()

    @property
    def is_valid(self) -> bool:
        """Get the token validity.

        True if essential token data is present and is not expired.
        """
        return (
            self.tenant is not None
            and self.subject is not None
            and self.domain is not None
            and not self.is_expired
        )

    def to_dict(self):
        """Get the main token attributes."""
        return dict(
            tenant=self.tenant,
            domain=self.domain,
            subject=self.subject,
            expires_at=str(self.expires_at),
            is_valid=self.is_valid
        )

    def __repr__(self) -> str:
        """Show the implementing class an main attributes."""
        return f'<{self.__class__.__name__}({json.dumps(self.to_dict())})>'

    def __str__(self) -> str:
        """Render the token string."""
        return self.token_string

    def __bool__(self) -> bool:
        """Get the validity of the token."""
        return self.is_valid


class WaylayTokenAuth(_http.Auth):
    """Authentication flow with a waylay token.

    Will automatically refresh an expired token.
    """

    current_token: Optional[WaylayToken]
    credentials: WaylayCredentials

    def __init__(
        self,
        credentials: WaylayCredentials,
        initial_token: Optional[TokenString] = None,
        credentials_callback: Optional[CredentialsCallback] = None
    ):
        """Create a Waylay Token authentication provider."""
        self.credentials = credentials
        self.current_token = None

        if isinstance(credentials, TokenCredentials):
            initial_token = initial_token or credentials.token

        if initial_token:
            try:
                self.current_token = self._create_and_validate_token(initial_token)
            except AuthError:
                pass

        self.credentials_callback = credentials_callback

    def auth_flow(self, request: _http.Request) -> Generator[_http.Request,  _http.Response, None]:
        """Authenticate a http request.

        Implements the authentication callback for the http client.
        """
        token = self.assure_valid_token()
        request.headers["Authorization"] = f"Bearer {token}"
        yield request

    def assure_valid_token(self) -> WaylayToken:
        """Validate the current token and request a new one if invalid."""
        if self.current_token:
            # token exists and is valid
            return self.current_token

        self.current_token = self._create_and_validate_token(self._request_token_string())
        return self.current_token

    def _create_and_validate_token(self, token: TokenString) -> WaylayToken:
        return WaylayToken(token).validate()

    def _request_token_string(self) -> TokenString:
        """Request a token."""
        if isinstance(self.credentials, NoCredentials):
            if self.credentials_callback is not None:
                # TODO: where is this used? Clients need to update their
                # definition of the callback to use gateway URL as argument.
                self.credentials = self.credentials_callback(
                    self.credentials.accounts_url or self.credentials.gateway_url
                )
            else:
                raise AuthError("No credentials or credentials_callback provided.")

        if isinstance(self.credentials, TokenCredentials):
            raise AuthError(
                f"cannot refresh expired token with credentials "
                f"of type '{self.credentials.credentials_type}'"
            )

        if isinstance(self.credentials, ApplicationCredentials):
            raise AuthError(
                f"credentials of type {self.credentials.credentials_type} are not supported yet"
            )

        if isinstance(self.credentials, ClientCredentials):
            return _request_token(self.credentials)

        raise AuthError(
            f"credentials of type {self.credentials.credentials_type} are not supported"
        )


def _request_token(credentials: ClientCredentials) -> str:
    token_url_prefix = credentials.accounts_url or f'{credentials.gateway_url}/accounts/v1'
    token_url = f"{token_url_prefix}/tokens?grant_type=client_credentials"
    token_req = {
        'clientId': credentials.api_key,
        'clientSecret': credentials.api_secret,
    }
    try:
        token_resp = _http.post(url=token_url, json=token_req)
        if token_resp.status_code != 200:
            raise AuthError(f'could not obtain waylay token: {token_resp.content!r} [{token_resp.status_code}]')
        token_resp_json = token_resp.json()
    except _http.HTTPError as exc:
        raise AuthError(f'could not obtain waylay token: {exc}') from exc
    return token_resp_json['token']


_AUTH_MESSAGE_FOR_EXCEPTON_CLASS = [
    (InvalidTokenError, 'invalid token'),
    (ExpiredSignatureError, 'token expired'),
    (PyJWTError, 'invalid token'),
    (TypeError, 'could not decode token'),
    (ValueError, 'could not decode token')
]


def _auth_message_for_exception(exception):
    for (exc_class, msg) in _AUTH_MESSAGE_FOR_EXCEPTON_CLASS:
        if isinstance(exception, exc_class):
            return msg
    return 'could not decode token'


def parse_credentials(json_obj: Dict[str, Any]) -> WaylayCredentials:
    """Convert a parsed json representation to a WaylayCredentials object."""
    cred_type = json_obj.get('type', None)
    if cred_type is None:
        raise ValueError('invalid json for credentials: missing type')

    for clz in [NoCredentials, ClientCredentials, ApplicationCredentials, TokenCredentials]:
        if clz.credentials_type == cred_type:  # type: ignore
            return clz(**{k: v for k, v in json_obj.items() if k != 'type'})

    raise ValueError(f'cannot parse json for credential type {cred_type}')



"""REST definitions for the 'runtime' entity of the 'byoml' service."""

from .._base import WaylayResource
from .._decorators import (
    return_body_decorator,
    return_path_decorator,
)
from ._decorators import (
    byoml_exception_decorator,
    byoml_retry_decorator,
)


RUNTIME_ARG = {
    'name': 'runtime',
    'type': 'str',
    'description': 'Name of a supported Byoml runtime.',
    'examples': ['byoml-sklearn-0.24']
}
RUNTIME_RETURN = {
    'name': 'runtime',
    'type': 'Dict',
    'description': 'Representation of a supported runtime.'
}
RUNTIME_LIST_RETURN = {
    'name': 'runtimes',
    'type': 'List[Dict]',
    'description': 'Representation of supported runtimes.'
}


class RuntimeResource(WaylayResource):
    """REST Resource for the 'runtime' entity of the 'byoml' service."""

    link_roots = {
        'doc': '${doc_url}/api/byoml/?id=',
        'apidoc': '${apidoc_url}/byoml.html'
    }

    actions = {
        'list': {
            'method': 'GET',
            'url': '/runtimes',
            'returns': [RUNTIME_LIST_RETURN],
            'decorators': [
                byoml_exception_decorator,
                byoml_retry_decorator,
                return_path_decorator(['runtimes'])
            ],
            'description': 'List runtimes (framework and framework version).',
            'links': {
                'doc': 'runtimes',
                'apidoc': '',
            },
        },
        'get': {
            'method': 'GET',
            'url': '/runtimes/{}',
            'arguments': [RUNTIME_ARG],
            'returns': [RUNTIME_RETURN],
            'decorators': [
                byoml_exception_decorator,
                byoml_retry_decorator,
                return_body_decorator,
            ],
            'description': 'Get a supported runtime.',
            'links': {
                'doc': 'runtimes',
                'apidoc': '',
            },
        }
    }



"""Cli support for the 'doc' command."""

from argparse import Namespace, ArgumentParser
from tabulate import tabulate
import json

from ..exceptions import ConfigError

from ..auth_interactive import ask_yes_no

from ..config import WaylayConfig, DEFAULT_PROFILE
from ..client import WaylayClient

COMMANDS = [LIST, SHOW, CREATE, REMOVE, DEFAULT, TOKEN] = ['list', 'show', 'create', 'remove', 'default', 'token']
DEFAULT_COMMAND = LIST


def init_parser(parser: ArgumentParser):
    """Initialize parser for the config command."""
    cmd_parsers = parser.add_subparsers(
        dest='config_cmd',
        description='Manage configuration profiles.'
    )
    cmd_parsers.add_parser(
        LIST,
        help='List config profiles.',
        description='List config profiles.'
    )
    show_cmd = cmd_parsers.add_parser(
        SHOW,
        help='Show an existing config profile.',
        description='Show an existing config profile.'
    )
    show_cmd.add_argument(
        'profiles', nargs='*',
        help='Specify one or more profile names. Defaults to the current profile.',
    )
    create_cmd = cmd_parsers.add_parser(
        CREATE,
        help='Create a new profile',
        description='Create a new profile'
    )
    create_cmd.add_argument(
        'new_profile', nargs='?',
        help='Specify a new profile name. Defaults to the current profile.'
    )
    remove_cmd = cmd_parsers.add_parser(
        REMOVE,
        help='Remove config profile.',
        description='Remove config profile.',
    )
    remove_cmd.add_argument(
        'profiles', nargs='+',
        help='Specify one or more profile names.'
    )
    default_cmd = cmd_parsers.add_parser(
        DEFAULT,
        help='Make a profile default.',
        description='Make a profile default.',
    )
    default_cmd.add_argument(
        'from_profile', nargs='?',
        help='Specify the profile you want to copy as default.'
    )
    token_cmd = cmd_parsers.add_parser(
        TOKEN,
        help='Fetch a token for the active profile.',
        description='Fetch a token for the active profile.',
    )
    token_cmd.add_argument(
        '--decode', dest='token_decode', default=False, action="store_true",
        help='Show the decode token data.'
    )
    return parser


def handle_config_list_cmd(args: Namespace):
    """Execute the config list command."""
    header = [
        "profile", "current", "credential", "id", "cluster"
    ]
    current_profile = args.profile or DEFAULT_PROFILE
    table = [
        [
            profile,
            '*' if profile == current_profile else '',
            config.credentials.credentials_type if config else '*invalid*',
            config.credentials.id if config else '*invalid*',
            config.gateway_url or config.accounts_url if config else '*invalid*'
        ]
        for profile in WaylayConfig.list_profiles()
        for config in [WaylayConfig.load(profile, interactive=False, skip_error=True)]
    ]
    print(tabulate(table, headers=header, tablefmt='github'))


def handle_config_show_cmd(args: Namespace):
    """Execute the config show command."""
    current_profile = args.profile or DEFAULT_PROFILE
    profiles = args.profiles or [current_profile]
    for profile in profiles:
        try:
            config = WaylayConfig.load(profile, interactive=False)
        except ConfigError as exc:
            print(f'ERR: {exc}')
        else:
            print(config.config_file_path(profile))
            print(json.dumps(config.to_dict(obfuscate=True), indent=2))


def handle_config_create_cmd(args: Namespace):
    """Execute the config show command."""
    profile = args.new_profile or args.profile or DEFAULT_PROFILE
    print(f"Creating configuration profile '{profile}'.")
    if profile in WaylayConfig.list_profiles():
        print(f"ERR: profile '{profile}' already exists.")
        return
    try:
        config = WaylayConfig.load(profile, interactive=True)
    except ConfigError as exc:
        print(f'ERR: {exc}')
    else:
        print(config.config_file_path(profile))
        print(json.dumps(config.to_dict(obfuscate=True), indent=2))


def handle_config_remove_cmd(args: Namespace):
    """Execute the config remove command."""
    profile_matchers = args.profiles
    to_remove = [
        profile for profile in WaylayConfig.list_profiles()
        if profile in profile_matchers
    ]
    if not to_remove:
        print('No matching profiles found.')
    for remove_profile in to_remove:
        if ask_yes_no(f"Removing profile '{remove_profile}'? (y/n): ", True):
            WaylayConfig.delete(remove_profile)
            print(f"Removed '{remove_profile}'")
        else:
            print(f"Skipped '{remove_profile}'")


def handle_config_default_cmd(args: Namespace):
    """Execute the config default command."""
    profile = args.from_profile or args.profile or DEFAULT_PROFILE
    if (profile == DEFAULT_PROFILE):
        print('Already default profile.')
        return
    config = WaylayConfig.load(profile)
    config.profile = DEFAULT_PROFILE
    config.save()
    print(f"Copied and saved profile '{profile}' as default to '{config.config_file_path()}'")


def handle_config_token_cmd(args: Namespace):
    """Execute the config default command."""
    profile = args.profile or DEFAULT_PROFILE
    token = WaylayConfig.load(profile).get_valid_token()
    if args.token_decode:
        print(json.dumps(token.token_data, indent=4))
    else:
        print(token.token_string)


COMMAND_HANDLERS = {
    LIST: handle_config_list_cmd,
    SHOW: handle_config_show_cmd,
    CREATE: handle_config_create_cmd,
    REMOVE: handle_config_remove_cmd,
    DEFAULT: handle_config_default_cmd,
    TOKEN: handle_config_token_cmd
}


def handle_cmd(client: WaylayClient, args: Namespace) -> bool:
    """Handle config cmd."""
    cmd = args.config_cmd or DEFAULT_COMMAND
    handler = COMMAND_HANDLERS.get(cmd)
    if handler:
        handler(args)
        return True
    return False




"""Cli support for the 'service' command."""

from argparse import ArgumentParser, Namespace
import sys
import json
from tabulate import tabulate
from ..client import WaylayClient
from ..service.util.info import DESC_LEVELS

COMMANDS = [LIST, DOC] = ['list', 'doc']


def init_srv_parser(parser: ArgumentParser):
    """Initialize parser for the service command."""
    cmd_parsers = parser.add_subparsers(dest='srv_cmd', description='Inspect service definitions.')
    cmd_list = cmd_parsers.add_parser(
        LIST, help='List services.', description='List all defined services.')
    init_doc_parser(cmd_parsers.add_parser(
        DOC,
        help='Generate an HTML SDK overview.',
        description='Generate an HTML SDK overview.'
    ))
    cmd_list.add_argument(
        '--json', dest='format_json', default=False, action="store_true",
        help='If set, list as json data.'
    )
    cmd_list.add_argument(
        '--format', dest='format', default='github',
        help='A format supported by [tabulate](https://pypi.org/project/tabulate).'
    )
    return parser


def init_doc_parser(parser: ArgumentParser):
    """Initialize parser for the doc command."""
    parser.add_argument(
        '-s', '--service', dest='doc_service', nargs='*',
        help='Filter services to document.', default=None
    )
    parser.add_argument(
        '-r', '--resource', dest='doc_resource', nargs='*',
        help='Filter resources to document.', default=None
    )
    parser.add_argument(
        '-l', '--link', dest='doc_link', nargs='*', help='Filter doc sites links.', default=None
    )
    parser.add_argument(
        '-d', '--doc_url', dest='doc_url',
        help='Set the root of the documentation site.', default='https://docs.waylay.io/#'
    )
    parser.add_argument(
        '-a', '--apidoc_url', dest='apidoc_url',
        help='Set the root of the api documentation site.', default='https://docs.waylay.io/openapi/public/redocly'
    )
    parser.add_argument(
        '--desc-level', dest='desc_lvl', nargs='*',
        help='Specify the included descriptions.',
        choices=DESC_LEVELS,
        default=None
    )
    return parser


def handle_srv_cmd(waylay: WaylayClient, args: Namespace) -> bool:
    """Execute a service command."""
    if args.srv_cmd is None:
        handle_srv_list_cmd(waylay, Namespace(format_json=False, profile=None, format='github'))
        return True
    if args.srv_cmd == LIST:
        handle_srv_list_cmd(waylay, args)
        return True
    if args.srv_cmd == DOC:
        handle_doc_cmd(waylay, args)
        return True

    return False


def handle_doc_cmd(client: WaylayClient, args: Namespace) -> bool:
    """Execute documentation generation."""
    client.config.set_local_settings(doc_url=args.doc_url, apidoc_url=args.apidoc_url)
    print(client.util.info.action_info_html(
        service=args.doc_service, resource=args.doc_resource, links=args.doc_link, desc_levels=args.desc_lvl
    ))
    return True


def handle_srv_list_cmd(waylay: WaylayClient, args: Namespace):
    """Execute the service list command."""
    if args.format_json:
        json.dump([
            {
                'key': srv.service_key,
                'root_url': srv.get_root_url(),
                'description': srv.description,
                'resources': [
                    r.as_dict()
                    for r in srv.resources
                ]
            }
            for srv in waylay.services
        ], sys.stdout)
        return
    profile = f'profile "{args.profile}"' if args.profile else 'default profile'
    header = [
        "key", f" url for {profile}", "description"
    ]
    table = [
        [srv.service_key, srv.get_root_url() or '', srv.description]
        for srv in waylay.services
    ]
    print(tabulate(table, headers=header, tablefmt=args.format))



"""Interactive authentication callback for client credentials."""
from typing import Optional
import logging
import re
import urllib.parse

from getpass import getpass
import httpx

from .auth import (
    DEFAULT_GATEWAY_URL,
    WaylayCredentials,
    ClientCredentials,
    DEFAULT_ACCOUNTS_URL,
    ACCOUNTS_USERS_ME_PATH,
    AuthError
)

_http = httpx

log = logging.getLogger(__name__)


def tell(message: str):
    """Show an interactive authentication message."""
    print(message)


def ask(prompt: str) -> str:
    """Prompt user for information."""
    return input(prompt)


def ask_secret(prompt: str) -> str:
    """Prompt user for credential information."""
    return getpass(prompt=prompt)


def ask_yes_no(prompt: str, default: Optional[bool] = None) -> bool:
    """Keep prompting the user until response starts with a true or false character."""
    while True:
        resp = ask(prompt)
        if not resp and default is not None:
            return default
        resp_char = resp.lower()[0]
        if resp_char in ('n', 'f'):
            return False
        if resp_char in ('t', 'y'):
            return True


def ask_gateway(default_gateway_url: str):
    """Ask for the gateway api in an interactive dialog."""
    gateway_url: str = _gateway_url_for(default_gateway_url)
    gateway_validated = False
    while not gateway_validated:
        tell(f'Proposed api gateway: {gateway_url}')
        tell('Please confirm, or specify a gateway by platform id, hostname or url.')
        tell("""Examples:
    'enterprise' (or 'api.waylay.io') for the Enterprise platform,
    'https://waylay-api.mycompany.com' as a custom endpoint url
        """)
        gateway_url = ask(
            f'> Press enter to confirm, or specify an alternate gateway [{gateway_url}]: '
        ) or gateway_url
        gateway_url = _gateway_url_for(gateway_url)
        try:
            gateway_status_resp = httpx.get(f'{gateway_url}{ACCOUNTS_USERS_ME_PATH}')
        except Exception as err:
            tell(f"Cannot connect to '{gateway_url}: {err}")
            gateway_url = _gateway_url_for(default_gateway_url)
        else:
            gateway_status = gateway_status_resp.status_code
            gateway_validated = gateway_status == 401
            if not gateway_validated:
                msg = ('Should require authentication' if gateway_status == 200 else gateway_status_resp.reason_phrase)
                tell(f"Not a gateway URL: '{gateway_url}': {msg}")
                gateway_url = _gateway_url_for(default_gateway_url)
            else:
                tell(f'Using gateway: {gateway_url}')
    return gateway_url


def request_migrate_to_gateway_interactive(profile, msg):
    """Asks to migrate to an api-gateway configuration."""
    tell(msg)
    tell(
        'NOTE: Migrating to an api-gateway will make this '
        'profile unusable for older waylay-py versions.'
    )
    ans = ask_yes_no(f"Migrate '{profile}' to use an api-gateway? [Y/n]", True)
    if not ans:
        tell(f"Not migrating configuration profile '{profile}'")
    return ans


def request_client_credentials_interactive(
    default_gateway_url: str = DEFAULT_GATEWAY_URL
) -> WaylayCredentials:
    """Asks interactively for client credentials.

    Default callback provider for an interactive WaylayConfig.
    """
    tell("Authenticating to the Waylay Platform")
    gateway_url = ask_gateway(default_gateway_url)
    tell("Please provide client credentials for the waylay data client.")
    credentials = ClientCredentials(api_key='', api_secret='', gateway_url=gateway_url)
    retry = 0
    while not credentials.is_well_formed() and retry < 3:
        api_key = ask(prompt='> apiKey : ').strip()
        api_secret = ask_secret(prompt='> apiSecret : ').strip()
        credentials = ClientCredentials(
            api_key=api_key, api_secret=api_secret, gateway_url=gateway_url
        )
        if not credentials.is_well_formed():
            retry += 1
            if retry >= 3:
                tell('Too many attempts, failing authentication')
                raise AuthError('Too many attempts, failing authentication')
            tell('Invalid apiKey or apiSecret, please retry')
    return credentials


def request_client_credentials_interactive_legacy(
    default_accounts_url: str = DEFAULT_ACCOUNTS_URL
) -> WaylayCredentials:
    """Asks interactively for client credentials.

    Default callback provider for an interactive WaylayConfig.
    """
    tell("Authenticating to the Waylay Platform")

    accounts_url: str = default_accounts_url
    acc_validated = False
    while not acc_validated:
        tell(f'Using authentication provider at [{accounts_url}]')
        accounts_url = ask(
            '> alternative endpoint (press enter to continue)?: '
        ) or accounts_url
        accounts_url = _root_url_for(accounts_url)
        accounts_status_resp = httpx.get(accounts_url)
        acc_validated = not accounts_status_resp.is_error
        if acc_validated:
            tell(f"Authenticating at '{accounts_status_resp.json()['name']}'")
        else:
            tell(f"Cannot connect to '{accounts_url}': {accounts_status_resp.reason_phrase}")

    tell("Please provide client credentials for the waylay data client.")
    credentials = ClientCredentials(api_key='', api_secret='', accounts_url=accounts_url)
    retry = 0
    while not credentials.is_well_formed() and retry < 3:
        api_key = ask(prompt='> apiKey : ').strip()
        api_secret = ask_secret(prompt='> apiSecret : ').strip()
        credentials = ClientCredentials(
            api_key=api_key, api_secret=api_secret, accounts_url=accounts_url
        )
        if not credentials.is_well_formed():
            retry += 1
            if retry >= 3:
                tell('Too many attempts, failing authentication')
                raise AuthError('Too many attempts, failing authentication')
            tell('Invalid apiKey or apiSecret, please retry')
    return credentials


def request_store_config_interactive(profile, save_callback):
    """Save interactively the storage of credentials as a profile."""
    if ask_yes_no(
        f'> Do you want to store these credentials with profile={profile}? [Y]: ', True
    ):
        save_location = save_callback()
        tell(
            f"Credential configuration stored as \n\t{save_location}\n"
            "Please make sure this file is treated securely.\n"
            "If compromised, _Revoke_ the api-key on the Waylay Console!"
        )


def _gateway_url_for(url_input: str):
    """Infer the gateway url from user input or existing accounts url."""
    url_input = (url_input or '').lower()
    if 'accounts-api' in url_input:
        url_input = url_input.replace('accounts-api', 'api')
    elif url_input in ('api', '', 'enterprise'):
        url_input = 'api.waylay.io'
    elif re.match(r'^[a-z0-9]+$', url_input):
        url_input = f'api-{url_input}.waylay.io'
    return _root_url_for(url_input)


def _root_url_for(host_or_url: str) -> str:
    scheme, loc, path, query, fragment = urllib.parse.urlsplit(host_or_url)

    if not scheme and not loc:
        # make sure any host name is converted in a https:// url
        scheme = 'https'
        loc = path
        path = ''

    if path.endswith('/'):
        # tenant settings root urls are without trailing slash
        log.warning('Trailing slashes trimmed: %s', host_or_url)
        path = path.rstrip('/')

    return urllib.parse.urlunsplit([scheme, loc, path, query, fragment])




"""REST client for the Waylay Byoml Service."""

from .._base import WaylayRESTService

from .model import ModelResource
from .runtime import RuntimeResource
from .framework import FrameworkResource
from .about import AboutResource


class ByomlService(WaylayRESTService):
    """REST client for the Waylay BYOML Service."""

    service_key = 'byoml'
    config_key = 'byoml'
    gateway_root_path = '/ml/v1'

    resource_definitions = {
        'model': ModelResource,
        'about': AboutResource,
        'framework': FrameworkResource,
        'runtime': RuntimeResource
    }

    model: ModelResource
    runtime: RuntimeResource
    framework: FrameworkResource
    about: AboutResource



"""REST client for the Waylay Platform."""

from collections import defaultdict
from typing import (
    Optional, TypeVar, List, Mapping, Type, Iterable, Dict
)
import logging
import sys
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from .service import (
    WaylayService,
    WaylayRESTService,
    WaylayServiceContext,
    ByomlService,
    TimeSeriesService,
    ResourcesService,
    StorageService,
    UtilService,
    ETLService,
    DataService,
    QueriesService,
    SERVICES
)

from .service.analytics import AnalyticsServiceLegacy
from .exceptions import ConfigError

from .config import (
    WaylayConfig,
    TenantSettings,
    DEFAULT_PROFILE,
    SERVICE_KEY_GATEWAY,
    SERVICE_KEY_ACCOUNTS
)
from .auth import (
    WaylayCredentials,
    ClientCredentials,
    TokenCredentials,
)

S = TypeVar('S', bound=WaylayService)
logger = logging.getLogger(__name__)


class WaylayClient():
    """REST client for the Waylay Platform."""

    analytics: AnalyticsServiceLegacy = AnalyticsServiceLegacy()
    byoml: ByomlService
    config: WaylayConfig
    timeseries: TimeSeriesService
    resources: ResourcesService
    storage: StorageService
    util: UtilService
    etl: ETLService
    data: DataService
    queries: QueriesService

    @classmethod
    def from_profile(
        cls, profile: str = DEFAULT_PROFILE,
        *, interactive=True, gateway_url=None
    ):
        """Create a WaylayClient named profile.

        Uses credentials from environment variables or a locally stored configuration.
        """
        return cls(WaylayConfig.load(
            profile, interactive=interactive, gateway_url=gateway_url
        ))

    @classmethod
    def from_client_credentials(
        cls, api_key: str, api_secret: str, *,
        gateway_url=None, accounts_url=None,
        settings: TenantSettings = None
    ):
        """Create a WaylayClient using the given client credentials."""
        credentials = ClientCredentials(
            api_key, api_secret, **_auth_urls(gateway_url, accounts_url, settings)
        )
        return cls.from_credentials(credentials, settings=settings)

    @classmethod
    def from_token(
        cls, token_string: str, *,
        gateway_url=None, accounts_url=None, settings: TenantSettings = None
    ):
        """Create a WaylayClient using a waylay token."""
        credentials = TokenCredentials(
            token_string, **_auth_urls(gateway_url, accounts_url, settings)
        )
        return cls.from_credentials(credentials, settings=settings)

    @classmethod
    def from_credentials(
        cls, credentials: WaylayCredentials, settings: TenantSettings = None
    ):
        """Create a WaylayClient using the given client credentials."""
        return cls(WaylayConfig(
            credentials,
            settings=settings
        ))

    def __init__(self, config: WaylayConfig):
        """Create a WaylayConfig instance."""
        self._services: List[WaylayService] = []
        self.config = config
        self.load_services()

    @property
    def services(self) -> List[WaylayService]:
        """Get the services that are available through this client."""
        return self._services

    @property
    def service_context(self) -> WaylayServiceContext:
        """Get the WaylayServiceContext view on this client."""
        return self

    def configure(self, config: WaylayConfig):
        """Update this client with the given configuration."""
        self.config = config
        for srv in self._services:
            srv.configure(self.config, context=self.service_context)

    def list_root_urls(self) -> Mapping[str, Optional[str]]:
        """List the currently configured root url for each of the registered REST services."""
        return {
            srv.config_key: srv.get_root_url()
            for srv in self._services if isinstance(srv, WaylayRESTService)
        }

    def __repr__(self):
        """Get a technical string representation of this instance."""
        return (
            f"<{self.__class__.__name__}("
            f"services=[{','.join(list(srv_class.service_key for srv_class in self.services))}],"
            f"config={self.config}"
            ")>"
        )

    def load_services(self):
        """Load all services that are registered with the `waylay_services` entry point."""
        registered_service_classes = [
            srv_class
            for entry_point in entry_points(group='waylay_services')
            for srv_class in entry_point.load()
        ]
        if not registered_service_classes:
            logger.warning(
                "The package %s "
                "seems not to be installed properly. "
                "If it was installed during this runtime session, "
                "please restart the runtime before continuing.",
                __package__
            )
            registered_service_classes = SERVICES
        self.register_service(*registered_service_classes)

    def register_service(self, *service_class: Type[S]) -> Iterable[S]:
        """Create and initialize one or more service of the given class.

        Replaces any existing with the same service_key.
        """
        new_services = [srv_class() for srv_class in service_class]
        new_plugin_priorities: Dict[str, int] = defaultdict(int)
        for srv in new_services:
            new_plugin_priorities[srv.service_key] = max(
                srv.plugin_priority, new_plugin_priorities[srv.service_key]
            )

        # delete existing services
        to_delete_service_index = [
            idx for idx, srv in enumerate(self._services)
            if (
                srv.service_key in new_plugin_priorities and
                srv.plugin_priority <= new_plugin_priorities[srv.service_key]
            )
        ]

        for idx in reversed(to_delete_service_index):
            # delete indexed entries in list from the back
            del self._services[idx]

        # change service list
        for srv in new_services:
            if srv.plugin_priority == new_plugin_priorities[srv.service_key]:
                self._services.append(srv)
                setattr(self, srv.service_key, srv)

        # reconfigure
        self.configure(self.config)
        return new_services

    # implements WaylayServiceContext protocol
    def get(self, service_class: Type[S]) -> Optional[S]:
        """Get the service instance for the provided class, if it is registered.

        Implements the `WaylayServiceContext.get` protocol.
        """
        for srv in self._services:
            if isinstance(srv, service_class):
                return srv
        return None

    def require(self, service_class: Type[S]) -> S:
        """Get the service instance for the given class or raise a ConfigError.

        Implements the `WaylayServiceContext.require` protocol.
        """
        srv = self.get(service_class)
        if srv is None:
            raise ConfigError(f"service {service_class.__name__} is not available.")
        return srv

    def list(self) -> List[WaylayService]:
        """List all registered Services.

        Implements the `WaylayServiceContext.list` protocol.
        """
        return list(self._services)


def _auth_urls(gateway_url=None, accounts_url=None, settings: TenantSettings = None):
    if settings:
        gateway_url = gateway_url or settings.get(SERVICE_KEY_GATEWAY)
        accounts_url = accounts_url or settings.get(SERVICE_KEY_ACCOUNTS)
    return {'gateway_url': gateway_url, 'accounts_url': accounts_url}



from contextlib import contextmanager
from io import BytesIO
from typing import (
    Iterator, Callable, List, Any, Union, Tuple, Dict, Optional, BinaryIO
)
from pathlib import Path
from tarfile import TarFile
from zipfile import ZipFile

import json
import os
import tempfile

from ._exceptions import (
    ByomlValidationError,
)


# type aliases for documentation purpose
PathLike = Union[str, os.PathLike]
PytorchModel = Any
TensorflowModel = Any
XgboostModel = Any
SklearnModel = Any
ByomlModel = Union[PytorchModel, TensorflowModel, XgboostModel, SklearnModel]
ModelSerializer = Callable[[PathLike, ByomlModel], PathLike]


def assert_dir_exists(dir_name: PathLike):
    """Raise error if the input directory does not exist."""
    if not os.path.exists(str(dir_name)):
        raise ByomlValidationError(f"The directory '{dir_name}' does not exist.")


def _serialize_torch(work_dir: PathLike, trained_model: PytorchModel) -> PathLike:
    """Serialize a pytorch model to a `model.pt` file."""
    # assuming a TorchScript model
    import torch  # pylint: disable=import-error

    model_file = f'{work_dir}/model.pt'
    model_script = torch.jit.script(trained_model)
    model_script.save(model_file)
    return model_file


def _serialize_joblib(work_dir: PathLike, trained_model: SklearnModel) -> PathLike:
    """Serialize a sklearn model to a `model.joblib` file."""
    import joblib  # pylint: disable=import-error

    model_file = f'{work_dir}/model.joblib'
    joblib.dump(trained_model, model_file)
    return model_file


def _serialize_dill(work_dir: PathLike, trained_model: Any) -> PathLike:
    """Serialize a custom model to a `model.dill` file."""
    import dill  # pylint: disable=import-error

    model_path = f'{work_dir}/model.dill'
    with open(model_path, 'wb') as model_file:
        dill.settings['recurse'] = True
        dill.dump(trained_model, model_file)
    return model_path


def _serialize_tf(work_dir: PathLike, trained_model: TensorflowModel) -> PathLike:
    """Serialize a tensorflow model to a model folder."""
    import tensorflow as tf  # pylint: disable=import-error

    tf.saved_model.save(trained_model, work_dir)
    return work_dir


def _serialize_bst(work_dir: PathLike, trained_model: XgboostModel) -> PathLike:
    """Serialize a xgboost model to a `model.bst` file."""
    if hasattr(trained_model, 'save_model'):
        model_file = f'{work_dir}/model.bst'
        trained_model.save_model(model_file)
        return model_file
    raise ByomlValidationError('Could not serialise this model: missing `save_model` method.')


SUPPORTED_FRAMEWORKS: Dict[str, ModelSerializer] = {
    "pytorch": _serialize_torch,
    "sklearn": _serialize_joblib,
    "tensorflow": _serialize_tf,
    "xgboost": _serialize_bst,
    "custom": _serialize_dill,
}


class ModelArchiveBuilder:
    """A context to build model archives."""

    work_dir_path: Path

    def __init__(self, work_dir: Optional[PathLike]):
        """Create a model archive build context."""
        self.work_dir = work_dir
        self.temp_dir = None

    def __enter__(self):
        """Enter the model archive build context."""
        if self.work_dir:
            self.work_dir_path = Path(self.work_dir)
        else:
            self.temp_dir = tempfile.TemporaryDirectory()
            self.work_dir_path = Path(self.temp_dir.__enter__())
        return self

    def __exit__(self, typ, value, traceback):
        """Exit the model archive build context."""
        if self.temp_dir:
            self.temp_dir.__exit__(typ, value, traceback)

    def serialize_model(
        self, trained_model: ByomlModel, framework: str
    ) -> PathLike:
        """Serialize a model to a given path."""
        framework_function = SUPPORTED_FRAMEWORKS.get(framework, None)
        if framework_function is not None:
            return framework_function(str(self.work_dir_path), trained_model)

        raise ByomlValidationError(
            f'Passing a model instance is not supported for this `{framework}` model, '
            'please provide the path to the saved model instead.'
        )


class ModelZipArchiveBuilder(ModelArchiveBuilder):
    """A context to build a zip model archives (legacy byoml api)."""

    def _get_files_to_zip(self, file_or_dir: PathLike) -> List[Tuple[str, str]]:
        """Get the filenames to zip and the name it should have in the zipfile."""
        file_or_dir = str(file_or_dir)
        if not os.path.isdir(file_or_dir):
            # single file
            zip_file_name = os.path.basename(file_or_dir)
            return [(file_or_dir, zip_file_name)]

        file_names: List[Tuple[str, str]] = []
        for root, _, files in os.walk(file_or_dir):
            for file_name in files:
                # the root will always contain the same suffix, which should not end up in the zip
                zip_root = root[len(file_or_dir):]
                zip_file_name = os.path.join(zip_root, file_name)

                file_path = os.path.join(root, file_name)

                file_names.append((file_path, zip_file_name))

        return file_names

    @contextmanager
    def save_model_in_dir(
        self, trained_model: Union[PathLike, ByomlModel], framework: str
    ):
        """Create the model zip file in a temporary buffer."""
        if not isinstance(trained_model, (str, os.PathLike)):
            file_name = self.serialize_model(trained_model, framework)
            files = self._get_files_to_zip(file_name)
        else:
            files = self._get_files_to_zip(trained_model)

        model_zip_buffer = BytesIO()
        with ZipFile(model_zip_buffer, 'w') as zipper:
            for file_name, zip_file_name in files:
                zipper.write(file_name, zip_file_name)

        yield model_zip_buffer
        model_zip_buffer.close()


class ModelPlugArchiveBuilder(ModelArchiveBuilder):
    """A context to build a model plug archive (openfaas)."""

    model_spec_path: Path
    model_path: Path
    requirements_path: Path
    lib_path: Path

    def add_model_spec(
        self,
        model_spec: Dict
    ):
        """Add a model specification."""
        model_spec_path = self.work_dir_path / 'model.json'
        with open(model_spec_path, 'wt') as model_spec_file:
            json.dump(model_spec, model_spec_file)
        self.model_spec_path = model_spec_path

    def add_model(
        self, trained_model: Union[PathLike, ByomlModel], framework: str
    ):
        """Add a serialized model file."""
        if isinstance(trained_model, (str, os.PathLike)):
            self.model_path = Path(trained_model)
        else:
            self.model_path = Path(
                self.serialize_model(trained_model, framework)
            )

    def add_requirements(
        self,
        requirements_file: Optional[PathLike],
        requirements: Optional[str]
    ):
        """Add requirements file."""
        if requirements_file:
            if requirements:
                raise AttributeError(
                    "cannot specify both 'requirements' and 'requirements_file'"
                )
            self.requirements_path = Path(requirements_file)
        else:
            with open(self.work_dir_path / 'requirements.txt', 'wt') as f:
                f.write(requirements or '')
            self.requirements_path = self.work_dir_path / 'requirements.txt'

    def add_lib(self, lib: Optional[PathLike]):
        """Add a lib folder."""
        if not lib:
            # create an empty dir
            self.lib_path = self.work_dir_path / 'lib'
            self.lib_path.mkdir()
        else:
            lib_path = Path(lib)
            if not lib_path.is_dir():
                raise AttributeError(
                    f"argument 'lib={lib}' must refer to an existing directory."
                )
            self.lib_path = lib_path

    @contextmanager
    def create_plug_tar_archive(self) -> Iterator[Tuple[BinaryIO, int]]:
        """Create a tar archive buffer."""
        assert self.model_spec_path
        assert self.model_path
        tar_bytes = BytesIO()
        with TarFile.open(fileobj=tar_bytes, mode='w:gz') as tar_file:
            # spec file
            tar_file.add(self.model_spec_path, arcname='model.json')
            # model file or dir
            tar_file.add(self.model_path, arcname='models/' + self.model_path.name)
            tar_file.add(self.requirements_path, arcname='requirements.txt')
            tar_file.add(self.lib_path, arcname='lib')

        tar_size = tar_bytes.tell()
        tar_bytes.seek(0)
        yield tar_bytes, tar_size
        tar_bytes.close()



"""REST definitions for the 'model' entity of the 'byoml' service."""

from contextlib import contextmanager

from functools import wraps
from typing import (
    Callable, List, Any, Union, Dict, Optional
)
import json
import pandas as pd
import logging

from waylay.service import WaylayRESTService
from .._base import WaylayResource
from .._decorators import (
    return_body_decorator,
    return_path_decorator,
    default_timeout_decorator,
    suppress_header_decorator,
    log_server_timing_decorator,
)
from ._decorators import (
    byoml_exception_decorator,
    byoml_retry_decorator,
    byoml_raise_not_ready_get,
    byoml_retry_upload_after_deletion_decorator,
    DEFAULT_BYOML_MODEL_TIMEOUT
)
from ...exceptions import (
    RestRequestError,
)
from ._model_archive import (
    ModelPlugArchiveBuilder,
    ModelZipArchiveBuilder,
    ByomlModel, PathLike
)

LOG = logging.getLogger(__name__)

DEFAULT_MODEL_UPLOAD_TIMEOUT = 120

LOG_SERVER_TIMING = logging.getLogger(__name__ + '.server-timing')
byoml_server_timing_decorator = log_server_timing_decorator(LOG_SERVER_TIMING, logging.INFO)

MODEL_NAME_ARG = {
    'name': 'model_name',
    'type': 'str',
    'description': 'name for the model',
    'example': 'my_prediction_model_001'
}
MODEL_RESULT = {
    'name': 'model',
    'type': 'Dict',
    'description': 'A representation of the model deployment.'
}
MODEL_INFERENCE_RESULT = {
    'name': 'results',
    'type': 'List',
    'description': 'The results of a model inference, normally a list of list of numerical data.'
}


def _input_data_as_list(input_data):
    if isinstance(input_data, list):
        if not input_data:
            # empty list
            return input_data

        if hasattr(input_data[0], 'tolist'):
            # list of numpy arrays
            return [d.tolist() for d in input_data]

        if isinstance(input_data[0], (pd.DataFrame, pd.Series)):
            # list of pandas
            return [d.values.tolist() for d in input_data]

        # list of (list of ...) value types?
        return input_data

    # pandas
    if isinstance(input_data, (pd.DataFrame, pd.Series)):
        return input_data.values.tolist()

    # numpy arrays
    if hasattr(input_data, 'tolist'):
        return input_data.tolist()

    raise RestRequestError(
        f'input data of unsupported type {type(input_data)}'
    )


def model_execution_request_decorator(action_method):
    """Decorate an action to prepare the execution of the model.

    Transforms any input data into a list, and provides it
    as `instances` in the request body.
    """
    @wraps(action_method)
    def wrapped(model_name, input_data=None, **kwargs):
        body = kwargs.pop('body', {})
        if 'instances' not in body:
            body = {
                'instances': _input_data_as_list(input_data),
                **body
            }
        return action_method(
            model_name,
            body=body,
            **kwargs
        )
    return wrapped


def _execute_model_decorators(response_key: str) -> List[Callable]:
    return [
        byoml_server_timing_decorator,
        byoml_exception_decorator,
        byoml_retry_decorator,
        model_execution_request_decorator,
        return_path_decorator(
            [response_key],
            default_response_constructor=None,
            response_constructor_advice="Use 'np.array' to return a numpy array, 'pd.DataFrame' for a pandas dataframe."
        )
    ]


class ModelResource(WaylayResource):
    """REST Resource for the 'model' entity of the 'byoml' service."""

    service: WaylayRESTService

    link_roots = {
        'doc': '${doc_url}/api/byoml/?id=',
        'apidoc': '${apidoc_url}/byoml.html'
    }

    actions = {
        'list': {
            'method': 'GET',
            'url': '/models',
            'returns': [
                {
                    'name': 'models',
                    'type': 'List[Dict]',
                    'description': 'A list of metadata objects for available models'
                }
            ],
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                return_path_decorator(['available_models'])
            ],
            'description': 'List the metadata of the deployed <em>BYOML Models</em>',
            'links': {
                'doc': 'overview-of-the-api',
                'apidoc': '',
            },
        },
        'list_names': {
            'method': 'GET',
            'url': '/models',
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                return_path_decorator(['available_models', 'name'])
            ],
            'returns': [
                {
                    'name': 'model_names',
                    'type': 'List[str]',
                    'description': 'A list of names of deployed models.'
                }
            ],
            'description': 'List the names of deployed <em>BYOML Models</em>',
            'links': {
                'doc': 'overview-of-the-api',
                'apidoc': '',
            },
        },
        '_create': {
            'method': 'POST',
            'url': '/models',
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                default_timeout_decorator(DEFAULT_MODEL_UPLOAD_TIMEOUT),
                byoml_retry_upload_after_deletion_decorator,
                return_body_decorator,
            ],
            'description': (
                'Build and create a new <em>BYOML Model</em> as specified in the request'
            ),
            'links': {
                'doc': 'how-to-upload-your-model',
                'apidoc': '',
            },
        },
        'upload': {
            'wrapped_actions': ['_create']
        },
        '_replace': {
            'method': 'PUT',
            'url': '/models/{}',
            'arguments': [MODEL_NAME_ARG],
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                default_timeout_decorator(DEFAULT_MODEL_UPLOAD_TIMEOUT),
                return_body_decorator,
            ],
            'description': 'Build and replace the named <em>BYOML Model</em>',
            'links': {
                'doc': 'overwriting-a-model',
                'apidoc': '',
            },
        },
        'replace': {
            'wrapped_actions': ['_replace']
        },
        'get': {
            'method': 'GET',
            'url': '/models/{}',
            'arguments': [MODEL_NAME_ARG],
            'returns': [MODEL_RESULT],
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                byoml_raise_not_ready_get,
                byoml_retry_decorator,
                return_body_decorator,
            ],
            'description': 'Fetch the metadata of the named <em>BYOML Model</em>',
            'links': {
                'doc': 'checking-out-your-model',
                'apidoc': '',
            },
        },
        'update': {
            'method': 'PATCH',
            'url': '/models/{}',
            'arguments': [
                MODEL_NAME_ARG,
                {
                    'name': 'body',
                    'type': 'Dict[str,str]',
                    'description': 'metadata attributes for the model',
                    'examples': ['{"description":"Updated"}']
                }
            ],
            'returns': [MODEL_RESULT],
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                byoml_retry_decorator,
                return_body_decorator,
            ],
            'description': (
                'Update the metadata of the named <em>BYOML Model</em>.\n'
                'Only metadata attributes can be modified.'
            ),
            'links': {
                'doc': 'update-metadata-for-a-model',
                'apidoc': '',
            },
        },
        # ## not yet supported on plug-registry
        # 'get_content': {
        #     'method': 'GET',
        #     'url': '/models/{}/content',
        #     'arguments': [
        #         MODEL_NAME_ARG
        #     ],
        #     'decorators': [
        #         byoml_server_timing_decorator,
        #         byoml_exception_decorator
        #     ],
        #     'description': 'Fetch the content of the named <em>BYOML Model</em>',
        #     'links': {
        #         'doc': 'checking-out-your-model'
        #     },
        # },
        'examples': {
            'method': 'GET',
            'url': '/models/{}/examples',
            'arguments': [
                MODEL_NAME_ARG
            ],
            'returns': [
                {
                    'name': 'examples',
                    'type': 'List[Dict]',
                    'description': 'A list of examples for the model'
                }
            ],
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                byoml_retry_decorator,
                return_path_decorator(['example_payloads'])
            ],
            'description': (
                'Fetch the <em>example request input</em> of the named <em>BYOML Model</em>'
            ),
            'links': {
                'doc': 'example-input',
                'apidoc': '',
            },
        },
        'predict': {
            'method': 'POST',
            'url': '/models/{}/predict',
            'arguments': [MODEL_NAME_ARG],
            'returns': [MODEL_INFERENCE_RESULT],
            'decorators': _execute_model_decorators('predictions'),
            'description': (
                'Execute the <em>predict</em> capability of the named <em>BYOML Model</em>'
            ),
            'links': {
                'doc': 'predictions',
                'apidoc': '',
            },
        },
        'regress': {
            'method': 'POST',
            'url': '/models/{}/regress',
            'arguments': [MODEL_NAME_ARG],
            'returns': [MODEL_INFERENCE_RESULT],
            'decorators': _execute_model_decorators('result'),
            'description': (
                'Execute the <em>regress</em> capability of the named  <em>BYOML Model</em>'
            ),
            'links': {
                'doc': 'predictions',
                'apidoc': '',
            },
        },
        'classify': {
            'method': 'POST',
            'url': '/models/{}/classify',
            'arguments': [MODEL_NAME_ARG],
            'returns': [MODEL_INFERENCE_RESULT],
            'decorators': _execute_model_decorators('result'),
            'description': (
                'Execute the <em>classification</em> capability of the named <em>BYOML Model</em>'
            ),
            'links': {
                'doc': 'predictions',
                'apidoc': '',
            },
        },
        'remove': {
            'method': 'DELETE',
            'url': '/models/{}',
            'arguments': [MODEL_NAME_ARG],
            'decorators': [
                byoml_server_timing_decorator,
                byoml_exception_decorator,
                return_body_decorator,
            ],
            'description': 'Remove the named <em>BYOML Model</em>',
            'links': {
                'doc': 'deleting-a-model',
                'apidoc': '',
            },
        },
    }

    def __init__(self, *args, **kwargs):
        """Create a ModelResource."""
        kwargs.pop('timeout', None)
        super().__init__(*args, timeout=DEFAULT_BYOML_MODEL_TIMEOUT, **kwargs)

    @contextmanager
    def _send_model_arguments(
        self, model_name: str, trained_model: Union[PathLike, ByomlModel],
        framework: str = "sklearn",
        framework_version: Optional[str] = None,
        metadata: Optional[Dict] = None,
        work_dir: Optional[PathLike] = None,
        **kwargs
    ):
        """Upload a binary model with given name, framework and description."""
        with ModelZipArchiveBuilder(work_dir) as model_builder:
            with model_builder.save_model_in_dir(trained_model, framework) as model_zip_buffer:
                yield {
                    'body': {
                        "name": model_name,
                        "framework": framework,
                        "framework_version": framework_version,
                        "metadata": json.dumps(metadata or {}),
                    },
                    'files': {
                        "file": ('model.zip', model_zip_buffer.getvalue())
                    },
                    **kwargs
                }

    @contextmanager
    def _send_model_plug_arguments(
        self, model_name: str,
        trained_model: Union[PathLike, ByomlModel],
        framework: str = "sklearn",
        framework_version: Optional[str] = None,
        metadata: Optional[Dict] = None,
        requirements_file: Optional[PathLike] = None,
        requirements: Optional[str] = None,
        lib: Optional[PathLike] = None,
        work_dir: Optional[PathLike] = None,
        **kwargs
    ):
        runtime = self.service.framework.find_runtime(framework, framework_version)
        with ModelPlugArchiveBuilder(work_dir) as builder:
            builder.add_model_spec({
                "name": model_name,
                "version": "0.0.1",
                "metadata": metadata,
                "runtime": runtime['name']
            })
            builder.add_model(trained_model, runtime.get('framework', framework))
            builder.add_requirements(requirements_file, requirements)
            builder.add_lib(lib)
            with builder.create_plug_tar_archive() as (tar_bytes, tar_size):
                headers = kwargs.pop('headers', {})
                headers['Content-Type'] = 'application/tar+gzip'
                headers['Content-Length'] = str(tar_size)
                yield {
                    'body': tar_bytes,  # TODO or 'content': tar_bytes ??
                    'headers': headers,
                    **kwargs
                }

    @suppress_header_decorator('Content-Type')
    def upload(
        self,
        model_name: str,
        trained_model: Union[PathLike, ByomlModel],
        framework: str = "sklearn",
        framework_version: Optional[str] = None,
        metadata: Optional[Dict] = None,
        description: Optional[str] = None,
        requirements_file: Optional[PathLike] = None,
        requirements: Optional[str] = None,
        lib: Optional[PathLike] = None,
        work_dir: Optional[PathLike] = None,
        **kwargs
    ) -> Any:
        """Upload a new machine learning model with given name, framework and description.

Arguments:
    model_name      The name of the model.
    trained_model   The model object (will be serialised to an zip archive before upload),
                    or a file path to the serialized model file or folder.
    framework       One of the supported frameworks (default 'sklearn').
    framework_version
                    A supported framework version, (default `None`,
                    which selects the default server side)
    metadata       User modifiable metadata. Standardised ones are:
                    * 'description' A description of the model.
                    * 'author' The attributed author of the model.
                    * 'title' Used as display title in user interfaces.
    description     Description of the model. (Deprecated: prefer `metadata.description`)
    requirements_file
                    A file containing custom python requirements (above that of the framework)
    requirements    Custom python requirements (above those of the framework).
                    Only one of 'requirements' and 'requirements_file' can be specified.
    lib             A file or folder with custom packages that can be referred to in
                    the requirements file as 'lib/<custom-requirement>'
    work_dir        Optional location of the working directory used to serialize the model.
                    If not specified, a temporary directory is used.
    (other args)    Passed onto the underlying REST request
        """
        if requirements or requirements_file or lib:
            # use the new plug format upload (openfaas only)
            with self._send_model_plug_arguments(
                model_name, trained_model,
                framework=framework,
                framework_version=framework_version,
                metadata={
                    'description': description,
                    **(metadata or {})
                },
                requirements_file=requirements_file,
                requirements=requirements,
                lib=lib,
                work_dir=work_dir,
                **kwargs
            ) as arguments:
                return self._create(**arguments)  # pylint: disable=no-member
        else:
            # legacy format (to be removed after move to openfaas byoml 1.4.0)
            with self._send_model_arguments(
                model_name, trained_model,
                framework=framework,
                framework_version=framework_version,
                metadata={
                    'description': description,
                    **(metadata or {})
                },
                work_dir=work_dir,
                **kwargs
            ) as arguments:
                return self._create(**arguments)  # pylint: disable=no-member

    @suppress_header_decorator('Content-Type')
    def replace(
        self,
        model_name: str,
        trained_model: Union[PathLike, ByomlModel],
        framework: str = "sklearn",
        framework_version: Optional[str] = None,
        metadata: Optional[Dict] = None,
        description: Optional[str] = None,
        work_dir: Optional[PathLike] = None,
        **kwargs
    ) -> Any:
        """Replace a machine learning model with given name, framework and description.

Arguments:
    model_name      The name of the model.
    trained_model   The model object (will be serialised to an zip archive before upload).
    framework       One of the supported frameworks (default 'sklearn').
    framework_version
                    A supported framework version, (default `None`,
                    which selects the default server side)
    metadata       User modifiable metadata. Standardised ones are:
                    * 'description' A description of the model.
                    * 'author' The attributed author of the model.
                    * 'title' Used as display title in user interfaces.
    description     Description of the model. (Deprecated: prefer `metadata.description`)
    work_dir        Optional location of the working directory used to serialize the model.
                    If not specified, a temporary directory is used.
    (other)         Passed onto the underlying REST request.
        """
        with self._send_model_arguments(
                model_name, trained_model,
                framework=framework,
                framework_version=framework_version,
                metadata={
                    'description': description,
                    **(metadata or {})
                },
                work_dir=work_dir,
                **kwargs
        ) as arguments:
            return self._replace(   # pylint: disable=no-member
                model_name,
                **arguments,
            )
